#==========================================================
# Tables --
#
#   creation and management of database tables
#
# TODO
#   * convert PgAcVar global to variable (Tables scope)
#==========================================================
#

namespace eval Tables {
    variable Win
}

#------------------------------------------------------------
# ::Tables::export --
#
#   Given an open window name, performs the export of results
#   in that window.
#
# Arguments:
#   wn_     name of a window
#
# Returns:
#   none
#------------------------------------------------------------
#
proc ::Tables::export {wn_} {

    global PgAcVar CurrentDB

    # setup the name for the temporary table
    set tmptbl "results"
    append tmptbl "_" [clock format [clock seconds] -format "%H%M%S"]

    # start the clock and open the import/export window
    setCursor CLOCK
    ::ImportExport::setup 0

    # set a default for the text file
    set ::ImportExport::tablename $tmptbl
    set ::ImportExport::filename $tmptbl.txt
    set ::ImportExport::wizard::tablecols $Win($wn_,colnames)

    # create the table
    if {$Win($wn_,isaquery) || !$Win($wn_,updatable)} {
        # if its a query or view then dont worry about the ctid column
        set sql "CREATE TEMP TABLE $tmptbl
                                AS $Win($wn_,query)"
    } else {
        set sql "CREATE TEMP TABLE $tmptbl
                                AS SELECT * FROM $Win($wn_,tablename)"
    }
    set pgres [wpg_exec $CurrentDB $sql]
    setCursor NORMAL

};# # end proc ::Tables::export

#----------------------------------------------------------
# ::Tables::introspect --
#
#   Given a tablename, returns the SQL needed to recreate it
#
# Arguments:
#   tablename_  name of a table to introspect
#   dbh_        an optional database handle
#
# Returns:
#   insql       the CREATE and INSERT statements to make this table
#----------------------------------------------------------
#
proc ::Tables::introspect {tablename_ {dbh_ ""}} {

    set insql [::Tables::clone $tablename_ $tablename_ $dbh_]

    return $insql

}; # end proc ::Tables::introspect


#----------------------------------------------------------
# ::Tables::clone --
#
#   Like introspect, only changes the tablename
#
# Arguments:
#   srctable_   the original table
#   desttable_  the clone table
#   dbh_        an optional database handle
#   inserts_    optional, 1 to include data, 0 default
#
# Returns:
#   insql       the INSERT statement to clone this table
#----------------------------------------------------------
#
proc ::Tables::clone {srctable_ desttable_ {dbh_ ""} {inserts_ 0}} {

    global CurrentDB

    if {[string match "" $dbh_]} {
        set dbh_ $CurrentDB
    }

    set insql ""
    set colnames ""

    # first we build the CREATE statement
    set insql "CREATE TABLE [::Database::quoteObject $desttable_] (\n"
    set recs [::Database::getTableInfo $srctable_ $dbh_]
    # this is pretty much the same as the column info pane
    # in the tables design mode
    foreach recl $recs {
        array set rec $recl
        set fsize $rec(attlen)
        set fsize1 $rec(atttypmod)
        set ftype $rec(typname)
        if { $fsize=="-1" && $fsize1!="-1" } {
            set fsize $rec(atttypmod)
            incr fsize -4
        }
        if { $fsize1=="-1" && $fsize=="-1" } {
            set fsize ""
        }
        if {$rec(attnotnull) == "t"} {
            set notnull "NOT NULL"
        } else {
            set notnull ""
        }
        if {$rec(attnum)>0} {
            append insql "\t\"" $rec(attname) "\" " $ftype
            if {[string length $fsize]>0
             && [lsearch [list varchar bpchar] $ftype]!=-1} {
                append insql "(" $fsize ")"
            }
            if {[string length $notnull]>0} {
                append insql " " $notnull
            }
            append insql ",\n"
        }
    }
    set insql [string trimright [string trim $insql] ,]
    append insql "\n);"

    # then we built INSERT statements for each row in the table
    if {$inserts_} {
        set sql "SELECT $colnames
                   FROM [::Database::quoteObject $srctable_]"
        wpg_select $dbh_ $sql rec {
            append insql "INSERT INTO $desttable_ ($colnames) VALUES ("
            foreach c [split $colnames ,] {
                # lets check for empty fields and replace them with NULLs
                if {[string length $rec($c)] == 0 } {
                    append insql "NULL,"
                } else {
                    append insql "'" [::Database::quoteSQL $rec($c)] "',"
                }
            }
            set insql [string trimright $insql ,]
            append insql ");\n"
        }
    }

    return $insql

}; # end proc ::Tables::clone


# ---------------------------------------------------------------
#
#   For tables design
#
# ---------------------------------------------------------------
#
proc ::Tables::design {tablename} {

    global PgAcVar CurrentDB

    if {$CurrentDB==""} return;
    set PgAcVar(tblinfo,tablename) $tablename
    Window show .pgaw:TableInfo
    wm title .pgaw:TableInfo "[intlmsg {Table information}] : $PgAcVar(tblinfo,tablename)"

}; # end proc ::Tables::design


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::addNewIndex {} {

    global PgAcVar

    set iflds [$PgAcVar(tblinfo,cols) curselection]
    if {$iflds==""} {
        showError [intlmsg "You have to select index fields!"]
        return
    }
    set ifldslist {}
    foreach i $iflds {
        lappend ifldslist "\"[lindex [$PgAcVar(tblinfo,cols) get $i] 0]\""
    }
    set PgAcVar(addindex,indexname) $PgAcVar(tblinfo,tablename)_[join $ifldslist _]
    # Replace the quotes with underlines
    regsub -all -- {"} $PgAcVar(addindex,indexname) {_} PgAcVar(addindex,indexname)
    # Replace the double underlines
    while {[regsub -all -- {__} $PgAcVar(addindex,indexname) {_} PgAcVar(addindex,indexname)]} {}
    # Replace the final underline
    regsub -all -- {_$} $PgAcVar(addindex,indexname) {} PgAcVar(addindex,indexname)
    set PgAcVar(addindex,indexfields) [join $ifldslist ,]
    Window show .pgaw:AddIndex
    wm transient .pgaw:AddIndex .pgaw:TableInfo

}; # end proc ::Tables::addNewIndex


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::deleteIndex {} {

    global PgAcVar

    set sel [$PgAcVar(tblinfo,indexes) curselection]
    if {$sel == ""} {
        showError [intlmsg "You have to select an index!"]
        return
    }
    if {[tk_messageBox -title [intlmsg Warning] -parent .pgaw:TableInfo \
            -message [format [intlmsg "You choose to delete index\n\n %s \n\nProceed?"] [lindex [$PgAcVar(tblinfo,indexes) get $sel] 0]] \
            -type yesno -default no]=="no"} {return}
    if {[sql_exec noquiet "DROP INDEX \"[lindex [$PgAcVar(tblinfo,indexes) get $sel] 0]\""]} {
        refreshInfo Indexes
    }

}; # end proc ::Tables::deleteIndex


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::clusterIndex {{mw_ ""}} {

    global PgAcVar

    set sel [$PgAcVar(tblinfo,indexes) curselection]
    if {$sel == ""} {
        # its not an index but a table that we are clustering
        #
        # this should move to the database namespace
        # because it is only available in PG 7.4 (maybe 7.3) or later
        #
        if {[tk_messageBox \
                    -title [intlmsg Warning] \
                    -parent .pgaw:TableInfo \
                    -message [format [intlmsg "You choose to cluster table\n\n %s \n\nProceed?"] [::Database::quoteObject $PgAcVar(tblinfo,tablename)]] \
                    -type yesno -default no]=="no"} {return}

        if {[sql_exec noquiet "CLUSTER [::Database::quoteObject $PgAcVar(tblinfo,tablename)]"]} {
            refreshInfo Indexes
        }
    } else {
        if {[tk_messageBox \
                    -title [intlmsg Warning] \
                    -parent .pgaw:TableInfo \
                    -message [format [intlmsg "You choose to cluster index\n\n %s \n\nAll other indices will be lost!\nProceed?"] [lindex [$PgAcVar(tblinfo,indexes) get $sel] 0]] \
                    -type yesno -default no]=="no"} {return}

        if {[sql_exec noquiet "CLUSTER \"[lindex [$PgAcVar(tblinfo,indexes) get $sel] 0]\" ON [::Database::quoteObject $PgAcVar(tblinfo,tablename)]"]} {
            refreshInfo Indexes
        }
    }
#<<<<<<< tables.tcl
#    set l [split $tablename .]
#    if {[llength $l] == 2} {
#        set s [lindex $l 0]
#        set t [lindex $l 1]
#        set Win($wn,query) "SELECT ctid,* FROM \"$s\".\"$t\""
#    } else {
#        set Win($wn,query) "select ctid,* from \"$tablename\""
#    }
        #set Win($wn,query) "SELECT ctid,* FROM [::Database::quoteObject $tablename]"

###    set Win($wn,ukey) [::Database::getUniqueKeys $::CurrentDB [::Database::quoteObject $tablename]]
###    set Win($wn,updatable) 1
###
###    if {[llength $Win($wn,ukey)] == 0} {
###        set Win($wn,updatable) 0
###    }
###    set Win($wn,query) "SELECT [join [linsert $Win($wn,ukey) end *] ,]
###                                 FROM [::Database::quoteObject $tablename]"
###
###    set Win($wn,isaquery) 0
###    initVariables $wn
###    refreshRecords $wn
###    catch {wm title $wn "$tablename"}
#=======
#>>>>>>> 1.36

}; # end proc ::Tables::clusterIndex


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::addNewColumn {} {

    global PgAcVar

    if {$PgAcVar(addfield,name)==""} {
        showError [intlmsg "Empty field name ?"]
        focus .pgaw:AddField.e1
        return
    }
    if {$PgAcVar(addfield,type)==""} {
        showError [intlmsg "No field type ?"]
        focus .pgaw:AddField.e2
        return
    }
    if {![sql_exec quiet "
        ALTER TABLE \"$PgAcVar(tblinfo,tablename)\"
         ADD COLUMN \"$PgAcVar(addfield,name)\" $PgAcVar(addfield,type)"]} {
        showError "[intlmsg {Cannot add column}]\n\n$PgAcVar(pgsql,errmsg)"
        return
    }
    Window destroy .pgaw:AddField
    sql_exec quiet "
        UPDATE pga_layout
           SET colnames=colnames || ' {$PgAcVar(addfield,name)}', colwidth=colwidth || ' 150',nrcols=nrcols+1
         WHERE tablename='$PgAcVar(tblinfo,tablename)'"
    refreshInfo Columns

}; # end proc ::Tables::addNewColumn


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::renameColumn {} {

    global PgAcVar CurrentDB

    if {[string length [string trim $PgAcVar(tblinfo,new_cn)]]==0} {
        showError [intlmsg "Field name not entered!"]
        return
    }
    set PgAcVar(tblinfo,new_cn) [string trim $PgAcVar(tblinfo,new_cn)]
    if {$PgAcVar(tblinfo,old_cn) == $PgAcVar(tblinfo,new_cn)} {
        showError [intlmsg "New name is the same as the old one!"]
        return
    }
    foreach col [$PgAcVar(tblinfo,cols) get 0 end] {
        if {[lindex $col 0]==$PgAcVar(tblinfo,new_cn)} {
            showError [format [intlmsg {Column name '%s' already exists in this table!}] $PgAcVar(tblinfo,new_cn)]
            return
        }
    }

    if {[sql_exec noquiet "
        ALTER TABLE \"$PgAcVar(tblinfo,tablename)\"
      RENAME COLUMN \"$PgAcVar(tblinfo,old_cn)\"
                 TO \"$PgAcVar(tblinfo,new_cn)\""]} {
        refreshInfo Columns
        Window destroy .pgaw:RenameField
        #
        # this should rebuild the index frame but it doesnt
        #
        #.pgaw:TableInfo.f2.fl.ilb insert end [::Database::getTableIndexes $PgAcVar(tblinfo,tablename)]
        #
        # showing permissions
        set temp $PgAcVar(tblinfo,permissions)
        regsub -- "^\{" $temp {} temp
        regsub -- "\}$" $temp {} temp
        regsub -all -- "\"" $temp {} temp
        foreach token [split $temp ,] {
            set oli [split $token =]
            set uname [lindex $oli 0]
            set rights [lindex $oli 1]
            if {$uname == ""} {set uname PUBLIC}
            set r_select " "
            set r_update " "
            set r_insert " "
            set r_rule   " "
            if {[string first r $rights] != -1} {set r_select x}
            if {[string first w $rights] != -1} {set r_update x}
            if {[string first a $rights] != -1} {set r_insert x}
            if {[string first R $rights] != -1} {set r_rule   x}
            #
            # changing the format of the following line can affect the loadPermissions procedure
            # see below
            # well this line doesnt work anymore
            #.pgaw:TableInfo.f3.plb insert end [format "%-23.23s %11s %11s %11s %11s" $uname $r_select $r_update $r_insert $r_rule]
        }
    }
}; # end proc ::Tables::renameColumn


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::loadPermissions {} {

    global PgAcVar

    set sel [$PgAcVar(tblinfo,perms) curselection]
    if {$sel == ""} {
        bell
        return
    }
    set uname [lindex [$PgAcVar(tblinfo,perms) get $sel] 0]
    Window show .pgaw:Permissions
    wm transient .pgaw:Permissions .pgaw:TableInfo
    set PgAcVar(permission,username) $uname

    foreach T {junk select update insert rule} P [$PgAcVar(tblinfo,perms) get $sel] {
        set PgAcVar(permission,$T) [expr {"X"==$P}]
    }

    focus .pgaw:Permissions.f1.ename

}; # end proc ::Tables::loadPermissions


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::savePermissions {} {

    global PgAcVar

    if {$PgAcVar(permission,username) == ""} {
        showError [intlmsg "User without name?"]
        return
    }
    if {$PgAcVar(permission,username)=="PUBLIC"} {
        set usrname PUBLIC
    } else {
        set usrname "\"$PgAcVar(permission,username)\""
    }
    set tname [::Database::quoteObject $PgAcVar(tblinfo,tablename)]
    sql_exec noquiet "REVOKE all on $tname from $usrname"

    foreach T {select insert update rule} {
        if {$PgAcVar(permission,$T)} {
            sql_exec noquiet "GRANT $T on $tname to $usrname"
        }
    }

    refreshInfo Permissions

}; # end proc ::Tables::savePermissions


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::refreshInfo {page} {

    global PgAcVar CurrentDB

    switch -- $page {
        "General" {    
            set recs [::Database::getTableInfo $PgAcVar(tblinfo,tablename)]
            foreach recl $recs {
                 array set rec $recl
                set PgAcVar(tblinfo,owner) $rec(usename)
                set PgAcVar(tblinfo,tableoid) $rec(oid)
                set PgAcVar(tblinfo,ownerid) $rec(usesysid)
                set PgAcVar(tblinfo,numtuples) $rec(reltuples)
                set PgAcVar(tblinfo,numpages) $rec(relpages)
                set PgAcVar(tblinfo,permissions) $rec(relacl)
                if {$rec(relhaspkey)=="t"} {
                    set PgAcVar(tblinfo,hasprimarykey) [intlmsg Yes]
                } else {
                    set PgAcVar(tblinfo,hasprimarykey) [intlmsg No]
                }
                if {$rec(relhasrules)=="t"} {
                    set PgAcVar(tblinfo,hasrules) [intlmsg Yes]
                } else {
                    set PgAcVar(tblinfo,hasrules) [intlmsg No]
                }
            }
        }
        "Columns" {
            $PgAcVar(tblinfo,cols) delete 0 end
            set recs [::Database::getTableInfo $PgAcVar(tblinfo,tablename)]
             foreach recl $recs {
                array set rec $recl
                set fsize $rec(attlen)
                set fsize1 $rec(atttypmod)
                set ftype $rec(typname)
                if { $fsize=="-1" && $fsize1!="-1" } {
                    set fsize $rec(atttypmod)
                    incr fsize -4
                }
                if { $fsize1=="-1" && $fsize=="-1" } {
                    set fsize ""
                }
                if {$rec(attnotnull) == "t"} {
                    set notnull "NOT NULL"
                } else {
                    set notnull {}
                }
                if {$rec(attnum)>0} {
                    $PgAcVar(tblinfo,cols) insert end [list $rec(attname) $ftype $fsize $notnull]
                }
            }
        }
        "Indexes" {
            $PgAcVar(tblinfo,indexes) delete 0 end
            foreach idxname [::Database::getTableIndexes $PgAcVar(tblinfo,tablename)] {
                set idxname [string trim $idxname \"]
                wpg_select $CurrentDB "SELECT pg_index.*, pg_class.oid
                                FROM [::Database::qualifySysTable pg_index],
                    [::Database::qualifySysTable pg_class]
                                WHERE pg_class.relname='$idxname' 
                                    AND pg_class.oid=pg_index.indexrelid" rec {
                    if {$rec(indisunique)=="t"} {
                        set isunique [intlmsg Yes]
                    } else {
                        set isunique [intlmsg No]
                    }
                    if {$rec(indisclustered)=="t"} {
                        set isclustered [intlmsg Yes]
                    } else {
                        set isclustered [intlmsg No]
                    }
                    set indexfields {}
                    foreach field $rec(indkey) {
                        if {$field!=0} {
                            wpg_select $CurrentDB "SELECT attname FROM pg_attribute WHERE attrelid=$PgAcVar(tblinfo,tableoid) AND attnum=$field" rec1 {
                                lappend indexfields $rec1(attname)
                            }
                        }
                    }
                    $PgAcVar(tblinfo,indexes) insert end [list $idxname [join $indexfields ,] $isunique $isclustered]
                }
            }
        }
        "Permissions" {
            set recs [::Database::getTableInfo $PgAcVar(tblinfo,tablename)]
            foreach recl $recs {
                array set rec $recl
                set PgAcVar(tblinfo,permissions) $rec(relacl)
            }
            $PgAcVar(tblinfo,perms) delete 0 end
            set temp [string map {\" ""} [lindex $PgAcVar(tblinfo,permissions) 0]]
            foreach token [split $temp ,] {
                set oli [split $token =]
                set uname [lindex $oli 0]
                set rights [lindex $oli 1]
                if {$uname == ""} {set uname PUBLIC}
                foreach P {r w a R} {
                    set r($P) ""
                    if {[string first $P $rights] != -1} {set r($P) X}
                }
                $PgAcVar(tblinfo,perms) insert end [list $uname $r(r) $r(w) $r(a) $r(R)]
            }
        }
    }

}; # end proc ::Tables::refreshInfo


# ---------------------------------------------------------------
#
#    For tables new/open/
#
# ---------------------------------------------------------------
proc ::Tables::new {} {

    PgAcVar:clean nt,*
    Window show .pgaw:NewTable
    focus .pgaw:NewTable.etabn

}; # end proc ::Tables::new


#----------------------------------------------------------
# ::Tables::open --
#
#   takes a selected table and displays its information in a
#   graphical display window suitable for editing
#
# Parameters:
#   tablename   name of the table to open
#   filter      if we are obscuring results, ie. a WHERE clause
#   order       what column we should be sorting by
#
# Modifies:
#   main window display
#----------------------------------------------------------
#
proc ::Tables::open {tablename {filter ""} {order ""}} {
    global CurrentDB
    variable Win

    # prep the window
    set wn [getNewWindowName]
    set Win($wn,parent) .pgaw:Tables:
    set Win($wn,istoplevel) 1
    set Win($wn,number) $Win(count)
    set Win($wn,tablename) $tablename
    createWindow $wn 1
    loadLayout $wn $tablename
    displayWindow $wn

    # prep the filtering and sorting
    foreach {a b} [split $order] break
    if {![info exists a]} {set a ""}
    if {![info exists b]} {set b ""}
    set Win($wn,sortfield) $a
    set Win($wn,sortdirection) $b
    set Win($wn,filter) $filter

    set qtab [::Database::quoteObject $tablename]
    set Win($wn,ukey) [::Database::getUniqueKeys $CurrentDB $qtab]
    set Win($wn,updatable) [llength $Win($wn,ukey)]

    set Win($wn,query) "
        SELECT [join [linsert $Win($wn,ukey) end *] ,]
          FROM [::Database::quoteObject $tablename]"

    set Win($wn,newrec_fields) {}
    set Win($wn,newrec_values) {}
    set Win($wn,modifymode) ""
    set Win($wn,isaquery) 0

    # finally, load the records
    refreshRecords $wn

    return $wn

}; # end proc ::Tables::open


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::get_tag_info {wn itemid prefix} {

    set taglist [$wn.c itemcget $itemid -tags]
    set i [lsearch -glob $taglist $prefix*]
    set thetag [lindex $taglist $i]

    return [string range $thetag 1 end]

}; # end proc ::Tables::get_tag_info


#----------------------------------------------------------
# ::Tables::deleteRecord --
#
#   Removes row(s) from a table.
#
# Parameters:
#   wn  window we are using to delete the row(s)
#
# Returns:
#   none
#
# Modifies:
#   a database table
#----------------------------------------------------------
#
proc ::Tables::deleteRecord {wn} {
    global CurrentDB
    variable Win

    if { !$Win($wn,updatable) } {
        return
    }

    set taglist [$Win($wn,tableview) curcellselection]
    set delcnt [llength $taglist]
    if { $delcnt == 0 } {
        return
    }

    # need to get the unique rows
    set Row [list]
    foreach idx $taglist {
        lappend Row [lindex [split $idx ,] 0]
    }
    set Row [lsort -unique $Row]

    set del [tk_messageBox \
        -title [intlmsg "FINAL WARNING"] \
        -icon question \
        -parent $wn \
        -message [intlmsg "Delete current record(s) ?"] \
        -type yesno \
        -default no]
    if { $del == "no" } {
            return
    }

    if { ![sql_exec noquiet "BEGIN TRANSACTION"] } {
        showError "[intlmsg {Can't get transaction}]"
        return
    }

    # check unique keys and add a clause for each
    set commit 1
    foreach row $Row {
        set sql "
            DELETE FROM [::Database::quoteObject $Win($wn,tablename)]
            WHERE "
        set wheresql [list]
        foreach uk $Win($wn,ukey) {
            set colidx [$Win($wn,tableview) columnindex col$uk]
            set val [$Win($wn,tableview) getcells $row,$colidx]
            set quk [::Database::quoteObject $uk]
            set qval [::Database::quoteSQL $val]
            lappend wheresql "$quk = '$qval'"
        }
        append sql [join $wheresql " AND "]
        if { ![sql_exec noquiet $sql] } {
            set commit 0
            break
        }
    }

    if { $commit } {
        sql_exec noquiet "COMMIT TRANSACTION"
        $Win($wn,tableview) delete $Row
    } else {
        sql_exec noquiet "ROLLBACK TRANSACTION"
    }

}; # end proc ::Tables::deleteRecord


#----------------------------------------------------------
# ::Tables::sortFromHeader --
#
#    Sorts a table based on the column that was selected
#
# Arguments:
#    win_   The window of the table display
#    idx_   The index of the column that was selected
#
# Results:
#    nothing return. The selected column gets sorted
#----------------------------------------------------------
#
proc ::Tables::sortFromHeader {win_ idx_} {
    variable Win

    set wn [winfo parent $win_]

    set Win($wn,sortfield) [$Win($wn,tableview) columncget $idx_ -title]

    if {[string match "desc" $Win($wn,sortdirection)]} {
        set Win($wn,sortdirection) asc
    } else {
        set Win($wn,sortdirection) desc
    }
    refreshRecords $wn

    return

}; # end proc ::Tables::sortFromHeader


#----------------------------------------------------------
# ::Tables::loadLayout --
#
#   We keep track of how we look at tables because chances
#   are we will want to look at this table again in the
#   future in the same way.
#
# Parameters:
#   wn          window name to setup layout within
#   layoutname  name of a previously save layout
#
# TODO
#   make layout preservation work again
#----------------------------------------------------------
#
proc ::Tables::loadLayout {wn layoutname} {
    global CurrentDB
    variable Win

    return

    setCursor CLOCK

    set Win($wn,layout_name) $layoutname
    catch {
        unset Win($wn,colcount) Win($wn,colnames) Win($wn,colwidth) Win($wn,coldetail)
    }
    set Win($wn,layout_found) 0
    set sql "
        SELECT *,oid
          FROM pga_layout
         WHERE tablename='$layoutname'
      ORDER BY oid DESC"
    set pgres [wpg_exec $CurrentDB $sql]
    set pgs [pg_result $pgres -status]

    if {$pgs == "PGRES_TUPLES_OK"} {
        set nrlay [pg_result $pgres -numTuples]
        if {$nrlay >= 1} {
            set layoutinfo [pg_result $pgres -getTuple 0]
            set Win($wn,colcount) [lindex $layoutinfo 1]
            set Win($wn,colnames)  [lindex $layoutinfo 2]
            set Win($wn,colwidth) [lindex $layoutinfo 3]
            set goodoid [lindex $layoutinfo 4]
            set Win($wn,layout_found) 1
        }
        if {$nrlay > 1} {
            showError "Multiple ($nrlay) layout info found\n\nPlease report the bug!"
            sql_exec quiet "DELETE FROM pga_layout WHERE (tablename='$Win($wn,tablename)') AND (oid<>$goodoid)"
        }
    }
    pg_result $pgres -clear

    return

}; # end proc ::Tables::loadLayout


#----------------------------------------------------------
# ::Tables::selectRecords --
#
#   Handles pulling the records from the database and
#   inserting them row by row into a display table, up to
#   the maximum preferred number of rows.
#
# Parameters:
#   wn      name of the window to fill
#   sql     query to fill it with
#
# Modifies:
#   a previously created display table window
#
# Returns:
#   none
#
# TODO:
#   make this an asynchronous pull from the database
#----------------------------------------------------------
#
proc ::Tables::selectRecords {wn sql} {
    global PgAcVar ; # needed for preferences (fonts, rows)
    global CurrentDB
    variable Win

    set prefrows $PgAcVar(pref,rows)
    set Win($wn,newrec_fields) {}
    set Win($wn,newrec_values) {}
    set Win($wn,leftcol) 0
    set Win($wn,leftoffset) 0
    set Win($wn,crtrow) {}
    set Win($wn,modifymode) ""
    set Win($wn,msg) [intlmsg "Accessing data. Please wait ..."]

    setCursor CLOCK
    catch {$Win($wn,btnclose) configure -state disabled}
    $Win($wn,tableview) delete 0 end

    set ok 1
    set err ""

    # compute how many records total in this query
    set pgres [wpg_exec $CurrentDB "SELECT COUNT(*) FROM ($sql) AS blah"]
    if { [set err [pg_result $pgres -error]] != "" } {
        showError $err {} $Win($wn,tableview)
        set ok 0
    } else {
        update idletasks
        set Win($wn,nrecs) [lindex [pg_result $pgres -getTuple 0] 0]
        pg_result $pgres -clear
    }

    if { $ok } {

        # we _must_ have a row to start on
        if { ![info exists Win($wn,currpage)] } {
            set Win($wn,currpage) 1
        }
        set Win($wn,rowpagestart) [expr {round(($Win($wn,currpage) - 1) * $prefrows)}]
        incr Win($wn,rowpagestart)
        if { $Win($wn,rowpagestart) > $Win($wn,nrecs) } {
            set Win($wn,rowpagestart) [expr {$Win($wn,nrecs) - $prefrows}]
        }
        if { $Win($wn,rowpagestart) == ""
          || $Win($wn,rowpagestart) <= 0} {
            set Win($wn,rowpagestart) 1
        }
        # but what row/record _should_ we start on for this page
        set firstrow [expr {$Win($wn,rowpagestart) - 1}]
        if { $firstrow > $Win($wn,nrecs) } {
            set firstrow [expr {$Win($wn,nrecs) - $prefrows}]
        }
        if { $firstrow < 0 } {
            set firstrow 0
        }
    }

    # open a cursor over the results that we will show on
    # this page
    if { $ok } {
        set ok [sql_exec noquiet "BEGIN"]
    }
    if { $ok } {
        set ok [sql_exec noquiet "DECLARE mycursor CURSOR FOR $sql"]
    }
    if { $ok } {
        set ok [sql_exec noquiet "MOVE FORWARD $firstrow IN mycursor"]
    }
    if { $ok } {
        set pgres [wpg_exec $CurrentDB "FETCH $prefrows IN mycursor"]
        if { [pg_result $pgres -error] != "" } {
            set ok 0
        }
    }
    if { $ok } {
        set ok [sql_exec noquiet "END"]
    }

    # something went wrong somewhere
    if { !$ok } {
        sql_exec quiet "END"
        setCursor DEFAULT
        # silly to open a result window if the query failed
        if { $Win($wn,isaquery) } {
            array unset Win $wn,*
            destroy $wn
        } else {
            set Win($wn,rowpagestart) ""
            set Win($wn,rowpagestop) ""
            set Win($wn,nrecs) ""
            catch {$Win($wn,btnclose) configure -state normal}
            catch {$Win($wn,btnprevpage) configure -state disabled}
            catch {$Win($wn,btnnextpage) configure -state disabled}
            catch {$Win($wn,btnfirstpage) configure -state disabled}
            catch {$Win($wn,btnlastpage) configure -state disabled}
            if { $err != "" } {
                set Win($wn,msg) $err
            } else {
                set Win($wn,msg) [intlmsg "ERROR"]
            }
        }
        return
    }

    # get some stats on the result set into variables
    set Win($wn,pagerecs) [pg_result $pgres -numTuples]
    set Win($wn,colcount) [pg_result $pgres -numAttrs]
    set Win($wn,colnames) [pg_result $pgres -attributes]
    set Win($wn,coldetail) [pg_result $pgres -lAttributes]

    # set up the display column headers
    set wincols [list]
    $Win($wn,tableview) configure -columns $wincols
    foreach attr $Win($wn,colnames) {
        lappend wincols 20 $attr left
    }
    $Win($wn,tableview) configure -columns $wincols

    # give all columns a name in case they move around
    # (although their title will always be correct)
    foreach attr $Win($wn,colnames) {
        set idx [lsearch $Win($wn,colnames) $attr]
        $Win($wn,tableview) columnconfigure $idx \
            -name col$attr
    }

    # if we can do updates, make all columns editable
    if { $Win($wn,updatable) } {
        foreach ctl [::Database::getColumnsTypesList $Win($wn,tablename)] {
            set col [lindex $ctl 0]
            set typ [lindex $ctl 1]
            set idx [lsearch $Win($wn,colnames) $col]
            set Win($wn,coltypes,$col) $typ
            if { $idx != -1 } {
                # TODO get fancy here with more widgets
                if { $typ == "bool" } {
                    $Win($wn,tableview) columnconfigure $idx \
                        -editable 1 \
                        -editwindow checkbutton \
                        -width 0
                } else {
                    $Win($wn,tableview) columnconfigure $idx \
                        -editable 1
                }
            }
        }
    }

    # denote which columns are unique keys
    foreach uk $Win($wn,ukey) {
        set col [lsearch $Win($wn,colnames) $uk]
        # it would be bad to allow edits on oid column
        if { $uk == "oid" } {
            $Win($wn,tableview) columnconfigure $col \
                -labelimage ::icon::services-16 \
                -width 10 \
                -editable false
        } else {
            $Win($wn,tableview) columnconfigure $col \
                -labelimage ::icon::services-16
        }
    }

    # denote which column we are sorting by
    if { $Win($wn,sortfield) != "" } {
        set col [lsearch $Win($wn,colnames) $Win($wn,sortfield)]
        if { $Win($wn,sortdirection) == "asc" } {
            $Win($wn,tableview) columnconfigure $col \
                -labelimage ::icon::1uparrow-16
        } else {
            $Win($wn,tableview) columnconfigure $col \
                -labelimage ::icon::1downarrow-16
        }
    }

    for {set i 0} {$i < $Win($wn,pagerecs)} {incr i} {
        set curtup [pg_result $pgres -getTuple $i]
        $Win($wn,tableview) insert end $curtup
        if { ![expr {$i % 10}] } {
            update idletasks
        }
    }
    set rowcount $i

    pg_result $pgres -clear


    # only allow sorts on columns in display
    catch {$Win($wn,cbsortfield) configure \
            -values $Win($wn,colnames)}

    set Win($wn,updatekey) oid
    set Win($wn,keylist) {}
    set Win($wn,msg) "Loading maximum $prefrows records ..."

    # compute the total number of pages to be used
    set Win($wn,lastpage) [expr {round(ceil($Win($wn,nrecs) / double($prefrows)))}]

    set Win($wn,last_rownum) $i
    set Win($wn,rowpagestop) [expr {$Win($wn,rowpagestart) + $rowcount - 1}]
    # no rows
    if { $Win($wn,rowpagestart) == 1
      && $Win($wn,rowpagestop) == 0
      && $Win($wn,nrecs) == 0} {
        set Win($wn,rowpagestart) ""
        set Win($wn,rowpagestop) ""
    }

    set Win($wn,toprec) 0
    set Win($wn,dirtyrec) 0

    set Win($wn,msg) [intlmsg "Ready."]

    # can we go back and forth any pages at all
    if { $Win($wn,currpage) == 1 } {
        catch {$Win($wn,btnprevpage) configure -state disabled}
        catch {$Win($wn,btnfirstpage) configure -state disabled}
    } else {
        catch {$Win($wn,btnprevpage) configure -state normal}
        catch {$Win($wn,btnfirstpage) configure -state normal}
    }
    if { $Win($wn,lastpage) <= 1 || $Win($wn,currpage) == $Win($wn,lastpage) } {
        catch {$Win($wn,btnnextpage) configure -state disabled}
        catch {$Win($wn,btnlastpage) configure -state disabled}
    } else {
        catch {$Win($wn,btnnextpage) configure -state normal}
        catch {$Win($wn,btnlastpage) configure -state normal}
    }

    catch {$Win($wn,btnclose) configure -state normal}
    setCursor DEFAULT

    return

}; # end proc ::Tables::selectRecords


#----------------------------------------------------------
# ::Tables::refreshRecords --
#
#   Updates the display in the given window.
#
# Parameters:
#   wn  name of the window to refresh
#
# Returns:
#   none
#
# Modfies:
#   graphical display of the records in the window wn
#----------------------------------------------------------
#
proc ::Tables::refreshRecords {wn} {
    variable Win

    # we will only apply filters and sorts to a scrap copy
    # of the query, this lets us change the filters and
    # sorts on the fly
    set qry $Win($wn,query)

    if { $Win($wn,isaquery) } {
        if { $Win($wn,filter) != "" || $Win($wn,sortfield) != ""} {
            showError [intlmsg "Sorting and Filtering not (yet) available from queries!\n\nPlease enter them in the query definition!"] queries $wn
            set Win($wn,sortfield) {}
            set Win($wn,filter) {}
        }
    } else {
        if {$Win($wn,filter) != ""} {
            append qry " WHERE ($Win($wn,filter))"
        }
        if {$Win($wn,sortfield) != ""} {
            append qry "
                ORDER BY [::Database::quoteObject $Win($wn,sortfield)]
                $Win($wn,sortdirection)"
        }
    }

    selectRecords $wn $qry

}; # end proc refreshRecords


#----------------------------------------------------------
# ::Tables::startRecordEdit --
#
#   Grabs a row so it can be edited interactively.
#
# Parameters:
#   win_    name of the tablelist widget
#   row_    row to edit
#   col_    column to edit
#   txt_    current data in the row/column chosen for edit
#
# Returns:
#   the text to place in the editing widget (usually an
#   entry box)
#
# Modifies:
#   state of the row in a database table
#
# TODO
#   lock the row for edit
#----------------------------------------------------------
#
proc ::Tables::startRecordEdit {win_ row_ col_ txt_} {
    variable Win

    set wn [winfo parent $win_]

    if { !$Win($wn,updatable) } {
        $Win($wn,tableview) cancelediting
        return
    }

    set Win($wn,dirtyrec) 1

    # check if we are UPDATEing or INSERTing
    # (no oid yet if we are inserting)
    # what to do when we dont have oids ???
    foreach uk $Win($wn,ukey) {
        set colidx [$Win($wn,tableview) columnindex col$uk]
        set val [$Win($wn,tableview) getcells $row_,$colidx]
        if { $val != "" } { break }
    }

    if { $val == "" } {
        set Win($wn,modifymode) "insert"
        set Win($wn,insertfailed) 0

    # they moved to some row that already has an oid
    # so lets finish up the insert on the last row on the screen
    # that does not yet have the oid
    } elseif { $Win($wn,modifymode) == "insert" } {
        insertRecord $win_
    }

    # we save the old value so we can check if it changed
    # when they are done editing the value
    if { $Win($wn,modifymode) != "insert" } {
        set Win($wn,modifymode) "update"
        set Win($wn,oldvalue) $txt_
    }

    # setup a binding on the automagic entry for double clicks
    # to open a large window for multiline editing
    set entrypath [$Win($wn,tableview) entrypath]
    # might not have a path, ie. it's a boolean checkbutton
    if { $entrypath != "" } {
        $entrypath configure -textvariable Win($wn,moretext)
        bind $entrypath <Double-1> {
            moreText [%W get] 1 [%W cget -textvariable]
        }
    }

    return $txt_

}; # end proc ::Tables::startRecordEdit


#----------------------------------------------------------
# ::Tables::endRecordEdit --
#
#   Updates the table when we are done editing a cell in a
#   row, or inserts a row.  Depends on what was being done.
#
# Parameters:
#   win_    name of the tablelist widget
#   row_    row to edit
#   col_    column to edit
#   txt_    current data in the row/column chosen for edit
#
# Returns:
#   the updated contents of the cell
#----------------------------------------------------------
#
proc ::Tables::endRecordEdit {win_ row_ col_ txt_} {
    variable Win

    set wn [winfo parent $win_]

    # remove binding for multi-line edit window
    set entrypath [$Win($wn,tableview) entrypath]
    if { $entrypath != "" } {
        bind $entrypath <Double-1> {}
    }

    if { $Win($wn,modifymode) == "update" } {
        return [::Tables::updateRecord $win_ $row_ $col_ $txt_]
    } elseif { $Win($wn,modifymode) == "insert" } {
        # always try to finish the insert if they tabbed off the
        # end of the row, and we had good input
        if { [$Win($wn,tableview) columnindex end] == $col_ } {
            insertRecord $win_ 1 $txt_
            set ec 0
            set cc [$Win($wn,tableview) columncount]
            for {set i 0} {$i < $cc} {incr i} {
                if { [$Win($wn,tableview) columncget $i -editable] } {
                    set ec $i
                    break
                }
            }
            if { $Win($wn,insertfailed) } {
                # is this cheating ???
                after 1 [list $::Tables::Win($wn,tableview) activate end]
                after 1 [list $::Tables::Win($wn,tableview) editcell end,$ec]
                # because this doesnt work:
                #$Win($wn,tableview) editcell end,1
            } else {
                after 1 [list $::Tables::Win($wn,tableview) cancelediting]
                after 1 [list $::Tables::Win($wn,tableview) activate end]
                after 1 [list $::Tables::Win($wn,tableview) seecell end,$ec]
            }
        }
        return $txt_
    } else {
        return
    }

}; # end proc ::Tables::endRecordEdit


#----------------------------------------------------------
# ::Tables::updateRecord --
#
#   Performs UPDATE of a displayed row into the database.
#
# Parameters:
#   win_    name of the tablelist widget
#   row_    row to edit
#   col_    column to edit
#   txt_    current data in the row/column chosen for edit
#
# Returns:
#   the updated contents of the cell
#
# Modifies:
#   state of the row in a database table
#----------------------------------------------------------
#
proc ::Tables::updateRecord {win_ row_ col_ txt_} {
    global CurrentDB
    variable Win

    set wn [winfo parent $win_]

    # certainly not
    if { !$Win($wn,updatable) } {
        return
    }

    # check if we even changed values
    if { $Win($wn,oldvalue) == $txt_ } {
        return $txt_
    }

    # prepare the sql
    set col [$Win($wn,tableview) columncget $col_ -title]
    set qcol [::Database::quoteObject $col]
    set qtxt [::Database::quoteSQL $txt_]
    set sql "
        UPDATE [::Database::quoteObject $Win($wn,tablename)]
        SET $qcol = '$qtxt'
        WHERE "

    # check unique keys and add a clause for each
    set wheresql [list]
    foreach uk $Win($wn,ukey) {
        set colidx [$Win($wn,tableview) columnindex col$uk]
        set val [$Win($wn,tableview) getcells $row_,$colidx]
        set quk [::Database::quoteObject $uk]
        set qval [::Database::quoteSQL $val]
        lappend wheresql "$quk = '$qval'"
    }
    append sql [join $wheresql " AND "]

    set Win($wn,msg) [intlmsg "Updating record ..."]

    # force the user to enter good data
    if { ![sql_exec noquiet $sql $CurrentDB $Win($wn,tableview)] } {
        $Win($wn,tableview) rowconfigure $row_ -bg red
        set Win($wn,msg) [intlmsg "Please re-enter the data."]
        $Win($wn,tableview) rejectinput
    } else {
        set Win($wn,modifymode) ""
        $Win($wn,tableview) rowconfigure $row_ -bg lightgreen
        after 1000 [list set ::Tables::Win($wn,msg) [intlmsg "Ready."]]
    }

    return $txt_

}; # end proc ::Tables::updateRecord


#----------------------------------------------------------
# ::Tables::insertRecord --
#
#   Performs INSERT of a record into a database table.
#
# Parameters:
#   win_    name of the tablelist widget
#   istxt_  1 if we have some text to insert for the last row
#   txt_    last column in a new row
#
# Modifies:
#   a row in a database table
#
# Remarks:
#  the istxt/txt is a little funky, but it works
#----------------------------------------------------------
#
proc ::Tables::insertRecord {win_ {istxt_ 0} {txt_ ""}} {
    global CurrentDB
    variable Win

    set wn [winfo parent $win_]

    # certainly not
    if { !$Win($wn,updatable) } {
        return
    }

    if { $Win($wn,insertfailed) } {
        set Win($wn,insertfailed) 0
        return
    }

    set Win($wn,msg) [intlmsg "Saving new record ..."]

    set row [$Win($wn,tableview) rowcget active -text]

    set hdrs [list]
    set vals [list]

    # types that cannot be empty strings
    set ntl [list timestamp timestamptz monetary numeric bool int2 int4 int8 float float4 float8]

    for {set i 0} {$i < [$Win($wn,tableview) columncount]} {incr i} {
        if { [$Win($wn,tableview) columncget $i -editable] } {
            set h [$Win($wn,tableview) columncget $i -title]
            set coltype $Win($wn,coltypes,$h)
            set v [lindex $row $i]
            # accept explicit null entries
            if { $v == "\\0" } {
                lappend vals "NULL"
                lappend hdrs [::Database::quoteObject $h]
            } elseif { $v == "" && [lsearch $ntl $coltype] != -1 } {
                # do nothing if we are a type that cannot
                # handle a text string
            } else {
                lappend vals "'[::Database::quoteSQL $v]'"
                lappend hdrs [::Database::quoteObject $h]
            }
        }
    }

    if { $istxt_ } {
        set vals [lreplace $vals end end "'[::Database::quoteSQL $txt_]'"]
    }

    set sql "
        INSERT INTO [::Database::quoteObject $Win($wn,tablename)]
                    ([join $hdrs ,])
             VALUES ([join $vals ,])"

    set pgres [wpg_exec $CurrentDB $sql]

    if { [set err [pg_result $pgres -error]] != "" } {
        $Win($wn,tableview) rowconfigure end -bg red
        set Win($wn,msg) [intlmsg "Please re-enter the row's data."]
        showError $err {} $Win($wn,tableview)
        set Win($wn,insertfailed) 1

    } else {
        # what gets returned on an INSERT when we have no oids ???
        if { [lsearch $Win($wn,ukey) "oid"] != -1 } {
            set oid [pg_result $pgres -oid]
            pg_result $pgres -clear
            set qry $Win($wn,query)
            append qry " WHERE oid=$oid"
            set pgres [wpg_exec $CurrentDB $qry]
            if { [set err [pg_result $pgres -error]] != "" } {
                showError $err {} $Win($wn,tableview)
            } else {
                set curtup [pg_result $pgres -getTuple 0]
                $Win($wn,tableview) delete end
                $Win($wn,tableview) insert end $curtup
            }
        }
        set Win($wn,modifymode) ""
        $Win($wn,tableview) rowconfigure end -bg lightgreen
        after 1000 [list set ::Tables::Win($wn,msg) [intlmsg "Ready."]]
    }
    pg_result $pgres -clear

}; # end proc ::Tables::insertRecord


#----------------------------------------------------------
# ::Tables::prepareNewRecord --
#
#   Makes a new row at the end of the current page which you
#   can edit interactively.
#
# Parmeters:
#   wn  tablelist we are inserting a row into
#   x   x-coordinate of the mouse (optional)
#   y   y-coordinate of the mouse (optional)
#
# Returns:
#   none
#
# Modifies:
#   a new row in the tablelist
#----------------------------------------------------------
#
proc ::Tables::prepareNewRecord {wn {x ""} {y ""}} {
    variable Win

    # certainly not
    if { !$Win($wn,updatable) } {
        return
    }

    # dont allow double inserts
    if { $Win($wn,modifymode) == "insert" } {
        return
    }

    set Win($wn,modifymode) "insert"
    set Win($wn,insertfailed) 0
    $Win($wn,tableview) insert end [list]

    # find the first column that we can edit
    set editcel -1
    set numcols [$Win($wn,tableview) columncount]
    for {set i 0} {$i < $numcols} {incr i} {
        if { [$Win($wn,tableview) columncget $i -editable] } {
            set editcel $i
            break
        }
    }

    # could be we can't edit any columns ???
    if { $editcel != -1 } {
        set Win($wn,msg) [intlmsg "Please enter the row's data."]
        $Win($wn,tableview) editcell end,$editcel
    }

}; # end proc ::Tables::prepareNewRecord


#----------------------------------------------------------
# ::Tables::getNewWindowName --
#
#   Keeps track of how many windows we have open in the
#   table view, and gives us a new one when we request it.
#
# Parameters:
#   none
#
# Returns:
#   an available new window name
#----------------------------------------------------------
#
proc ::Tables::getNewWindowName {} {
    variable Win

    if { ![info exists Win(count)] } {
        set Win(count) [clock seconds]
    }

    return .pgaw:Tables:[incr Win(count)]

}; # end proc ::Tables::getNewWindowName


#------------------------------------------------------------
# closeWin --
#
#    Closes the table view window.
#
# Arguments:
#    wn    the window path
#
# Results:
#    none
#
# Modifies:
#   frees stuff from the Win array
#------------------------------------------------------------
#
proc ::Tables::closeWin {wn} {
    variable Win
    wm withdraw $wn
    array unset Win $wn,*
    return
}; # end proc ::Tables::closeWin


#----------------------------------------------------------
# ::Tables::displayWindow --
#
#   Actually pops up the created window, if it were a
#   toplevel.
#
# Parameters:
#   win_    name of window to display
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Tables::displayWindow {win_} {

    if { [winfo exists $win_] } {
        wm deiconify $win_
    }

}; # end proc ::Tables::displayWindow


#----------------------------------------------------------
# ::Tables::changePage --
#
#   Changes to a different page of results.
#
# Parameters:
#   win_    name of window we are changing results for
#   move_   what page to move to: first, prev, next, last
#
# Returns:
#   none
#
# Modifies:
#   Win array used for keeping track of what page we are on,
#   also all the results in the window are refreshed
#----------------------------------------------------------
#
proc ::Tables::changePage {win_ move_} {
    variable Win

    set curr $Win($win_,currpage)

    switch $move_ {
        first   { set Win($win_,currpage) 1 }
        prev    { incr Win($win_,currpage) -1 }
        next    { incr Win($win_,currpage) 1 }
        last    { set Win($win_,currpage) $Win($win_,lastpage) }
    }

    if { $Win($win_,currpage) <= 1 } {
        set Win($win_,currpage) 1
    }

    if { $Win($win_,currpage) >= $Win($win_,lastpage) } {
        set Win($win_,currpage) $Win($win_,lastpage)
    }

    if { $curr != $Win($win_,currpage) } {
        refreshRecords $win_
    }

}; # end proc ::Tables::changePage


#----------------------------------------------------------
# ::Tables::createWindow --
#
#   Creates a new table view window to show the contents
#   of a table or the results of a query
#
# Parameters:
#   win_        the window we are creating, probably a name
#               returned from getNewWindowName
#   toplevel_   default is 0, set to 1 for a top level window
#
# Returns:
#   the widget base path, which should be the same as win_
#----------------------------------------------------------
#
proc ::Tables::createWindow {win_ {toplevel_ 0}} {
    global PgAcVar ; # for font prefs
    variable Win

    set base $win_

    if { $toplevel_ } {
        if {[winfo exists $base]} {
            wm withdraw $base
            return
        }
        toplevel $base -class TableView
        wm withdraw $base
        wm geometry $base 650x400
        if { ![info exists Win($base,tablename)] } {
            wm title $base "Table"
        } else {
            wm title $base [intlmsg "Table $Win($base,tablename)"]
        }
    }

    set Win($base,filter) {}
    set Win($base,sortfield) {}
    set Win($base,sortdirection) asc

    bind $base <Key-Insert> [list ::Tables::prepareNewRecord $base %x %y]
    bind $base <Control-Delete> [list ::Tables::deleteRecord $base]
    bind $base <Key-F1> [list Help::load tables]
    bind $base <Key-Prior> [list Tables::changePage $base prev]
    bind $base <Key-Next> [list Tables::changePage $base next]

    # only put the window dressing on a toplevel
    if { $toplevel_ } {

        frame $base.f1 \
            -borderwidth 1

        label $base.f1.l1  -borderwidth 0 -text [intlmsg {Sort field}]

        set Win($base,cbsortfield) [ComboBox $base.f1.e1 \
            -editable 0 \
            -width 14  \
            -modifycmd [list Tables::refreshRecords $base] \
            -textvariable Tables::Win($base,sortfield)]

        $base.f1.e1 bind <Key-Return>   [list Tables::refreshRecords $base]
        $base.f1.e1 bind <Key-KP_Enter> [list Tables::refreshRecords $base]

        label $base.f1.l2  \
            -text [intlmsg {Filter conditions}]
        entry $base.f1.e2  \
            -textvariable Tables::Win($base,filter)

        bind $base.f1.e2 <Key-Return> [list Tables::refreshRecords $base]
        bind $base.f1.e2 <Key-KP_Enter> [list Tables::refreshRecords $base]

        Button $base.f1.b3  \
            -borderwidth 1 \
            -image ::icon::up-22 \
            -helptext "Export Results" \
            -relief link \
            -command [list ::Tables::export $base]

        set Win($base,btnclose) [Button $base.f1.b1  \
            -borderwidth 1 \
            -image ::icon::exit-22 \
            -helptext "Close Window" \
            -relief link \
            -command [list ::Tables::closeWin $base]]

        Button $base.f1.b2  \
            -borderwidth 1 \
            -image ::icon::reload-22 \
            -helptext "Reload" \
            -relief link \
            -command [list ::Tables::refreshRecords $base]
    }

    frame $base.bottomfm \
        -borderwidth 1 \
        -height 75

    set Win($base,btnfirstpage) [Button $base.bottomfm.10  \
        -borderwidth 1 \
        -image ::icon::2leftarrow-22 \
        -relief link \
        -helptext [intlmsg "First Page"] \
        -command [list Tables::changePage $base first]]

    set Win($base,btnprevpage) [Button $base.bottomfm.11  \
        -borderwidth 1 \
        -image ::icon::1leftarrow-22 \
        -relief link \
        -helptext [intlmsg "Prev Page"] \
        -command [list Tables::changePage $base prev]]

    label $base.bottomfm.12 \
        -anchor w \
        -borderwidth 1 \
        -height 1  \
        -relief flat \
        -foreground #225588 \
        -text {} \
        -textvariable Tables::Win($base,msg)

    label $base.bottomfm.13 \
        -anchor w \
        -borderwidth 1 \
        -height 1  \
        -relief flat \
        -foreground #225588 \
        -text {}

    Entry $base.bottomfm.14 \
        -width 6 \
        -borderwidth 1 \
        -relief flat \
        -editable false \
        -helptext [intlmsg "First row on this page"] \
        -textvariable Tables::Win($base,rowpagestart)

    # dont let the user change this, its just a pain
    #bind $base.bottomfm.14 <Key-Return> [list Tables::refreshRecords $base]
    #bind $base.bottomfm.14 <Key-KP_Enter> [list Tables::refreshRecords $base]

    label $base.bottomfm.15 \
        -borderwidth 1 \
        -height 1  \
        -relief flat \
        -foreground #225588 \
        -text "-"

    Entry $base.bottomfm.16 \
        -width 6 \
        -borderwidth 1 \
        -relief flat \
        -editable false \
        -helptext [intlmsg "Last row on this page"] \
        -textvariable Tables::Win($base,rowpagestop)

    # dont let the user change this, its just a pain
    #bind $base.bottomfm.16 <Key-Return> [list Tables::refreshRecords $base]
    #bind $base.bottomfm.16 <Key-KP_Enter> [list Tables::refreshRecords $base]

    label $base.bottomfm.17 \
        -borderwidth 1 \
        -height 1  \
        -relief flat \
        -foreground #225588 \
        -text "/"

    Entry $base.bottomfm.18 \
        -width 6 \
        -borderwidth 1 \
        -relief flat \
        -editable false \
        -helptext [intlmsg "Total rows"] \
        -textvariable Tables::Win($base,nrecs)

    set Win($base,btnnextpage) [Button $base.bottomfm.19 \
            -borderwidth 1 \
            -image ::icon::1rightarrow-22 \
            -relief link \
            -helptext [intlmsg "Next Page"] \
            -command [list Tables::changePage $base next]]

    set Win($base,btnlastpage) [Button $base.bottomfm.20 \
            -borderwidth 1 \
            -image ::icon::2rightarrow-22 \
            -relief link \
            -helptext [intlmsg "Last Page"] \
            -command [list Tables::changePage $base last]]

    set Win($base,tableview) [tablelist::tablelist $base.c \
        -font $PgAcVar(pref,font_fix) \
        -selectmode extended \
        -stretch all \
        -takefocus 1 \
        -selectbackground #DDDDDD \
        -selectforeground #708090 \
        -showarrow 1 \
        -labelcommand Tables::sortFromHeader \
        -editstartcommand Tables::startRecordEdit \
        -editendcommand Tables::endRecordEdit \
        -forceeditendcommand 1 \
        -yscrollcommand [list $base.vsb set] \
        -xscrollcommand [list $base.hsb set]
    ]

    scrollbar $base.vsb  \
        -orient vert \
        -command [list $Win($base,tableview) yview]

    scrollbar $base.hsb  \
        -orient horizontal \
        -command [list $Win($base,tableview) xview]

    set row 0
    if { $toplevel_ } {
        incr row

        grid $base.f1 \
            -sticky ew \
            -columnspan 2

        pack $base.f1.l1 -side left -padx 4
        pack $base.f1.e1 -side left -padx 4
        pack $base.f1.l2 -side left -padx 4
        pack $base.f1.e2 -side left -padx 4
        pack $base.f1.b3 -side right
        pack $base.f1.b1 -side right
        pack $base.f1.b2 -side right
    }

    grid $base.c \
        -sticky news \
        -row $row \
        -column 0

    grid $base.vsb \
        -sticky ns \
        -row $row \
        -column 1

    grid $base.hsb \
        -sticky ew \
        -column 0 \
        -row [expr {$row + 1}]

    grid $base.bottomfm \
        -sticky ew \
        -columnspan 2 \
        -row [expr {$row + 2}]

    grid columnconfigure $base 0 -weight 1
    grid rowconfigure $base 1 -weight 1

    set col 0
    for {set i 10} {$i <= 20} {incr i} {
        grid $base.bottomfm.$i \
            -sticky we \
            -column $col \
            -row 0
        incr col
    }
    grid columnconfigure $base.bottomfm 2 -weight 10

    return $base

}; # end proc ::Tables::createWindow



##
##
##
proc colorize {num} {
    #if {$num == 1} {return totals}
    if {$num>0 && ($num%2 != 1)} { return colored }
}


##
##
##
proc colConfig {num} {

    #if {$num >= 2} {return right}

    return left

    return right
}; # end proc colConfig



#----------------------------------------------------------
# doesnt look we are using this proc
# since we have another one of the same name
# is this one old or is it new ?
#----------------------------------------------------------
#
proc ::Tables::renameColumn_DUPLICATE {} {

    global PgAcVar CurrentDB

    if {[string length [string trim $PgAcVar(tblinfo,new_cn)]]==0} {
        showError [intlmsg "Field name not entered!"]
        return
    }
    set old_name [string trim [string range $PgAcVar(tblinfo,old_cn) 0 31]]
    set PgAcVar(tblinfo,new_cn) [string trim $PgAcVar(tblinfo,new_cn)]
    if {$old_name == $PgAcVar(tblinfo,new_cn)} {
        showError [intlmsg "New name is the same as the old one!"]
        return
    }
    foreach line [.pgaw:TableInfo.f1.lb get 0 end] {
        if {[string trim [string range $line 0 31]]==$PgAcVar(tblinfo,new_cn)} {
            showError [format [intlmsg {Column name '%s' already exists in this table!}] $PgAcVar(tblinfo,new_cn)]
            return
        }
    }
    if {[sql_exec noquiet "alter table [::Database::quoteObject $PgAcVar(tblinfo,tablename)] rename column \"$old_name\" to \"$PgAcVar(tblinfo,new_cn)\""]} {
        refreshInfo Columns
        Window destroy .pgaw:RenameField
    }

}; # end proc ::Tables::renameColumn


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::createNewIndex {} {

    global PgAcVar

    if {$PgAcVar(addindex,indexname)==""} {
        showError [intlmsg "Index name cannot be null!"]
        return
    }

    setCursor CLOCK

    if {[sql_exec noquiet "CREATE $PgAcVar(addindex,unique) INDEX \"$PgAcVar(addindex,indexname)\" on [::Database::quoteObject $PgAcVar(tblinfo,tablename)] ($PgAcVar(addindex,indexfields))"]} {
        setCursor DEFAULT
        Window destroy .pgaw:AddIndex
        refreshInfo Columns
    }

    setCursor DEFAULT

}; # end proc ::Tables::createNewIndex


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::showIndexInformation {} {

    global PgAcVar CurrentDB

    set cs [.pgaw:TableInfo.f2.fl.ilb curselection]
    if {$cs==""} return

    set idxname [.pgaw:TableInfo.f2.fl.ilb get $cs]

    wpg_select $CurrentDB "
        SELECT pg_index.*,pg_class.oid
          FROM pg_index,pg_class
         WHERE pg_class.relname='$idxname'
           AND pg_class.oid=pg_index.indexrelid" rec {

        if {$rec(indisunique)=="t"} {
            set PgAcVar(tblinfo,isunique) [intlmsg Yes]
        } else {
            set PgAcVar(tblinfo,isunique) [intlmsg No]
        }

        if {$rec(indisclustered)=="t"} {
            set PgAcVar(tblinfo,isclustered) [intlmsg Yes]
        } else {
            set PgAcVar(tblinfo,isclustered) [intlmsg No]
        }

        set PgAcVar(tblinfo,indexfields) {}
        .pgaw:TableInfo.f2.fr.lb delete 0 end

        foreach field $rec(indkey) {
            if {$field!=0} {
##               wpg_select $CurrentDB "select attname from pg_attribute where attrelid=$PgAcVar(tblinfo,tableoid) and attnum=$field" rec1 {
##                   set PgAcVar(tblinfo,indexfields) "$PgAcVar(tblinfo,indexfields) $rec1(attname)"
##               }
                set PgAcVar(tblinfo,indexfields) "$PgAcVar(tblinfo,indexfields) $PgAcVar(tblinfo,f$field)"
                .pgaw:TableInfo.f2.fr.lb insert end $PgAcVar(tblinfo,f$field)
            }

        }
    }

    set PgAcVar(tblinfo,indexfields) [string trim $PgAcVar(tblinfo,indexfields)]

}; # end proc ::Tables::showIndexInformation


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::addNewColumn {} {

    global PgAcVar

    if {$PgAcVar(addfield,name)==""} {
        showError [intlmsg "Empty field name ?"]
        focus .pgaw:AddField.e1
        return
    }

    if {$PgAcVar(addfield,type)==""} {
        showError [intlmsg "No field type ?"]
        focus .pgaw:AddField.e2
        return
    }

    if {![sql_exec quiet "
        ALTER TABLE [::Database::quoteObject $PgAcVar(tblinfo,tablename)]
         ADD COLUMN \"$PgAcVar(addfield,name)\" $PgAcVar(addfield,type)"]} {
        showError "[intlmsg {Cannot add column}]\n\n$PgAcVar(pgsql,errmsg)"
        return
    }

    Window destroy .pgaw:AddField
    sql_exec quiet "
        UPDATE pga_layout
           SET colnames=colnames || ' {$PgAcVar(addfield,name)}', colwidth=colwidth || ' 150',nrcols=nrcols+1
         WHERE tablename='$PgAcVar(tblinfo,tablename)'"

    refreshInfo Columns

}; # end proc ::Tables::addNewColumn


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::newtable:add_new_field {} {

    global PgAcVar

    if {$PgAcVar(nt,fieldname)==""} {
        showError [intlmsg "Enter a field name"]
        focus .pgaw:NewTable.e2
        return
    }

    if {$PgAcVar(nt,fldtype)==""} {
        showError [intlmsg "The field type is not specified!"]
        return
    }

    if {($PgAcVar(nt,fldtype)=="varchar")&&($PgAcVar(nt,fldsize)=="")} {
        focus .pgaw:NewTable.e3
        showError [intlmsg "You must specify field size!"]
        return
    }

    set sup ""
    if {[lsearch {varbit varchar char decimal} $PgAcVar(nt,fldtype)] >= 0} {
        if {[string length $PgAcVar(nt,fldsize)] == 0} {
            showError [intlmsg "You must specify field size!"]
            return
        }
        set sup "($PgAcVar(nt,fldsize))"
    }

    if {[regexp -- $PgAcVar(nt,fldtype) "varchartextdatetime"]} {
        set supc "'"
    } else {
        set supc ""
    }

    # Don't put the ' arround default value if it contains the now() function
    if {([regexp -- $PgAcVar(nt,fldtype) "datetime"]) \
        && ([regexp -- now $PgAcVar(nt,defaultval)])} {
        set supc ""
    }

    # Clear the notnull attribute if field type is serial
    if {$PgAcVar(nt,fldtype)=="serial"} {
        set PgAcVar(nt,notnull) " "
    }

    if {$PgAcVar(nt,defaultval)==""} {
        set sup2 ""
    } else {
        set sup2 " DEFAULT $supc$PgAcVar(nt,defaultval)$supc"
    }

    # Checking for field name collision
    set inspos end

    set i 0
    foreach L [.pgaw:NewTable.cols get 0 end] {
        set linie [lindex $L 0]
        if {$PgAcVar(nt,fieldname)==[string trim $linie]} {
            if {[tk_messageBox \
                    -title [intlmsg Warning] \
                    -parent .pgaw:NewTable \
                    -message [format [intlmsg "There is another field with the same name: '%s'!\n\nReplace it ?"] $PgAcVar(nt,fieldname)] \
                    -type yesno \
                    -default yes]=="no"} return
            .pgaw:NewTable.cols delete $i
            set inspos $i
            break
        }
        incr i
    }

    .pgaw:NewTable.cols insert $inspos [list $PgAcVar(nt,fieldname) $PgAcVar(nt,fldtype)$sup $PgAcVar(nt,primarykey) $sup2$PgAcVar(nt,notnull)]

    focus .pgaw:NewTable.e2

    set PgAcVar(nt,fieldname) {}
    set PgAcVar(nt,fldsize) {}
    set PgAcVar(nt,defaultval) {}
    set PgAcVar(nt,primarykey) " "

    return

}; # end proc ::Tables::newtable:add_new_field


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::delete_field {} {

    set cur [.pgaw:NewTable.cols curselection]

    if {[string length $cur] == 0} {return}

    .pgaw:NewTable.cols delete $cur

    return

}; # end proc ::Tables::delete_field

#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Tables::newtable:create {} {

    global PgAcVar CurrentDB

    if {$PgAcVar(nt,tablename)==""} {
        showError [intlmsg "You must supply a name for your table!"]
        focus .pgaw:NewTable.etabn
        return
    }

    if {([.pgaw:NewTable.cols size]==0) && ($PgAcVar(nt,inherits)=="")} {
        showError [intlmsg "Your table has no columns!"]
        focus .pgaw:NewTable.e2
        return
    }

    set fl {}
    set pkf {}

    foreach line [.pgaw:NewTable.cols get 0 end] {
        set fldname "\"[string trim [lindex $line 0]]\""
        lappend fl "$fldname [lindex $line 1] [lindex $line 3]"
        if {[lindex $line 2]=="*"} {
            lappend pkf "$fldname"
        }
    }

    set temp "CREATE TABLE [::Database::quoteObject $PgAcVar(nt,tablename)] ([join $fl ",\n"]"

    if {$PgAcVar(nt,constraint)!=""} {
        set temp "$temp, constraint \"$PgAcVar(nt,constraint)\""
    }

    if {$PgAcVar(nt,check)!=""} {
        set temp "$temp check ($PgAcVar(nt,check))"
    }

    if {[llength $pkf]>0} {
        set temp "$temp, primary key([join $pkf ,])"
    }

    set temp "$temp)"

    if {$PgAcVar(nt,inherits)!=""} {
        set temp "$temp inherits ($PgAcVar(nt,inherits))"
    }

    setCursor CLOCK

    if {[sql_exec noquiet $temp]} {
        Window destroy .pgaw:NewTable
        Mainlib::cmd_Tables
    }

    setCursor DEFAULT

    return

}; # end proc ::Tables::newtable:create {} {


####################   END OF NAMESPACE TABLES ####################
####################   BEGIN VisualTcl CODE ####################



#----------------------------------------------------------
#----------------------------------------------------------
#
proc vTclWindow.pgaw:NewTable {base} {
global PgAcVar
    if {$base == ""} {
        set base .pgaw:NewTable
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    toplevel $base -class Toplevel
    wm focusmodel $base passive
    #wm geometry $base 634x392+78+181
    #wm maxsize $base 1280 1024
    #wm minsize $base 1 1
    wm overrideredirect $base 0
    #wm resizable $base 0 0
    wm deiconify $base
    wm title $base [intlmsg "Create new table"]

    # some helpful key bindings
    bind $base <Control-Key-w> [list destroy $base]

    bind $base <Key-F1> "Help::load new_table"

    label $base.l93 \
        -anchor w \
        -text [intlmsg {Table name}]

    entry $base.etabn \
        -textvariable PgAcVar(nt,tablename) 

    bind $base.etabn <Key-Return> [list focus .pgaw:NewTable.einh]

    label $base.li \
        -anchor w \
        -text [intlmsg Inherits]

    ComboBox $base.einh \
        -textvariable PgAcVar(nt,inherits) \
        -values [lsort [::Database::getTablesList]]

    bind $base.einh <Key-Return> [list focus .pgaw:NewTable.e2]

    label $base.lco \
        -anchor w \
        -text [intlmsg Constraint]

    entry $base.eco \
        -textvariable PgAcVar(nt,constraint) 

    label $base.lch \
        -text [intlmsg check]
    entry $base.ech \
        -textvariable PgAcVar(nt,check) 

    Separator $base.ll \
        -orient horizontal \
        -relief ridge 

    label $base.lab2 \
        -anchor w \
        -text [intlmsg {field name}]

    entry $base.e2 \
        -textvariable PgAcVar(nt,fieldname) 

    bind $base.e2 <Key-Return> [list focus .pgaw:NewTable.e1]

    label $base.lab1 \
        -text [intlmsg type]

    ComboBox $base.e1 \
        -textvariable PgAcVar(nt,fldtype) \
        -values [lsort [list \
                 char varchar text int2 int4 \
                 serial float4 float8 money abstime \
                 date datetime interval reltime time \
                 timespan timestamp boolean box circle \
                 line lseg path point polygon]]

    bind $base.e1 <Key-Return> [list focus .pgaw:NewTable.e5]

    label $base.lab4 \
        -anchor w \
        -text [intlmsg {Default value}]

    entry $base.e5 \
        -textvariable PgAcVar(nt,defaultval) 

    bind $base.e5 <Key-Return> [list focus .pgaw:NewTable.cb1]

    label $base.lab3 \
        -text [intlmsg size]

    entry $base.e3 \
        -textvariable PgAcVar(nt,fldsize) 

    bind $base.e3 <Key-Return> [list focus .pgaw:NewTable.e5]

    checkbutton $base.cb1 \
        -offvalue { } \
        -onvalue { NOT NULL} \
        -text [intlmsg {field cannot be null}] \
        -variable PgAcVar(nt,notnull) 


    checkbutton $base.pk \
        -offvalue { } \
        -onvalue * \
        -text [intlmsg {primary key}] \
        -variable PgAcVar(nt,primarykey) 


    #Button $base.addfld \
        #-relief link \
        #-command Tables::newtable:add_new_field \
        #-image ::icon::fileopen-22 \
        #-helptext [intlmsg "Add Field"]

    ButtonBox $base.bbox1 -spacing 2

    $base.bbox1 add \
        -image ::icon::down-22 \
        -padx 1 \
        -pady 1 \
        -relief link \
        -command [list ::Tables::newtable:add_new_field] \
        -helptext [intlmsg "Add Field"]

    $base.bbox1 add \
        -image ::icon::editdelete-22 \
        -padx 1 \
        -pady 1 \
        -relief link \
        -command [list ::Tables::delete_field] \
        -helptext [intlmsg "Delete Selected Field"]

    ButtonBox $base.bbox2 -spacing 2


    $base.bbox2 add \
        -image ::icon::gear-22 \
        -padx 1 \
        -pady 1 \
        -relief link \
        -command [list ::Tables::newtable:create] \
        -helptext [intlmsg "Create"]

    $base.bbox2 add \
        -image ::icon::cancel-22 \
        -padx 1 \
        -pady 1 \
        -relief link \
        -command {Window destroy .pgaw:NewTable} \
        -helptext [intlmsg "Cancel"]

    $base.bbox2 add \
        -image ::icon::help-22 \
        -padx 1 \
        -pady 1 \
        -relief link \
        -command [list ::Help::load new_table] \
        -helptext [intlmsg "Help"]

    scrollbar $base.yscroll \
        -command [list $base.cols yview]


    scrollbar $base.xscroll \
        -command [list $base.cols yview] \
        -orient horizontal

    tablelist::tablelist $base.cols \
        -yscrollcommand [list $base.yscroll set] \
        -xscrollcommand [list $base.xscroll set] \
        -labelcommand tablelist::sortByColumn \
        -selectbackground #DDDDDD \
        -selectmode extended \
        -font $PgAcVar(pref,font_normal) \
        -labelfont PGfont \
        -stretch all \
        -columns [list \
            0 [intlmsg "field name"] left \
            0 [intlmsg "type"] left \
            0 [intlmsg "pkey"] left \
            0 [intlmsg "options"] left] \
        -selectforeground #708090



    grid $base.l93 $base.etabn $base.li $base.einh -sticky w
    #grid configure $base.li $base.einh -sticky e
    grid $base.lco $base.eco $base.lch $base.ech -sticky w
    #grid configure $base.lch $base.ech -sticky e
    grid $base.ll -sticky ew -columnspan 6 -pady 4
    grid $base.lab2 $base.e2 $base.lab1 $base.e1 -sticky w
    grid $base.lab4 $base.e5 $base.lab3 $base.e3 -sticky w
    grid $base.cb1 $base.pk -sticky w -columnspan 2
    #grid configure $base.cb1 -columnspan 2
    grid $base.bbox1 x x $base.bbox2 -sticky w
    grid configure $base.bbox2 -sticky e
    grid $base.cols $base.yscroll -sticky news -columnspan 6
    grid configure $base.yscroll -sticky ns -columnspan 1
    grid $base.xscroll -sticky ew -columnspan 6


    grid columnconfigure $base {0 1 2 3 4 5} -weight 1
    grid rowconfigure $base 7 -weight 1


    return
}

proc vTclWindow.pgaw:TableInfo {base} {
global PgAcVar
    if {$base == ""} {
        set base .pgaw:TableInfo
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    toplevel $base -class Toplevel \
        -background #c7c3c7 

    #::tk::PlaceWindow $base widget .

    wm focusmodel $base passive
#    wm geometry $base 500x400+152+135
    wm minsize $base 100 100
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm deiconify $base
    wm title $base [intlmsg "Table information"]

    # some helpful key bindings
    bind $base <Control-Key-w> [list destroy $base]

    bind $base <Key-F1> "Help::load view_table_structure"
    NoteBook $base.nb
    pack $base.nb -expand 1 -fill both
    foreach i {General Columns Indexes Permissions} {
        $base.nb insert end $i \
            -text [intlmsg $i] \
            -raisecmd [list Tables::refreshInfo $i]
    }
    set page_general [$base.nb getframe General]
    set page_columns [$base.nb getframe Columns] 
    set page_indexes [$base.nb getframe Indexes]
    set page_permissions [$base.nb getframe Permissions]

    # general page
    TitleFrame $page_general.tf_id -text [intlmsg Identification] \
        -font $PgAcVar(pref,font_bold)
    pack $page_general.tf_id -side top -anchor nw

    set titleframe [$page_general.tf_id getframe]
    foreach {desc name} {
            {Table name} {tablename}
            {Table oid} {tableoid}
            {Owner} {owner}
            {Owner ID} {ownerid}
            {Has primary key?} {hasprimarykey}
            {Has rules?} {hasrules}
        } {
        label $titleframe.n$name -text [intlmsg $desc]

        entry $titleframe.v$name \
            -textvariable PgAcVar(tblinfo,$name) \
            -justify left \
            -state disabled \
            -width 30

        grid $titleframe.n$name $titleframe.v$name -sticky e
    }

    TitleFrame $page_general.tf_stat -text [intlmsg Statistics] \
        -font $PgAcVar(pref,font_bold)
    pack $page_general.tf_stat -side top -anchor nw

    set titleframe [$page_general.tf_stat getframe]
    foreach {desc name} {
            {Number of tuples} {numtuples}
            {Number of pages} {numpages}
        } {
        label $titleframe.n$name -text [intlmsg $desc]

        entry $titleframe.v$name \
            -textvariable PgAcVar(tblinfo,$name) \
            -justify left \
            -state disabled \
            -width 30

        grid $titleframe.n$name $titleframe.v$name -sticky e
    }
    
    # columns page
    scrollbar $page_columns.xscroll \
        -command [list $page_columns.cols xview] \
        -orient horizontal

    scrollbar $page_columns.yscroll \
        -command [list $page_columns.cols yview]

    set PgAcVar(tblinfo,cols) [tablelist::tablelist $page_columns.cols \
        -yscrollcommand [list $page_columns.yscroll set] \
        -xscrollcommand [list $page_columns.xscroll set] \
        -labelcommand tablelist::sortByColumn \
        -selectbackground #DDDDDD \
        -selectmode extended \
        -font $PgAcVar(pref,font_normal) \
        -labelfont PGfont \
        -stretch all \
        -columns [list \
            0 [intlmsg "field name"] left \
            0 [intlmsg "type"] left \
            0 [intlmsg "size"] left \
            0 [intlmsg "not null"] left] \
        -selectforeground #708090
    ]
    frame $page_columns.buttons
    button $page_columns.buttons.addcolbtn \
        -borderwidth 1 \
        -command {Window show .pgaw:AddField
            set PgAcVar(addfield,name) {}
            set PgAcVar(addfield,type) {}
            wm transient .pgaw:AddField .pgaw:TableInfo
            focus .pgaw:AddField.e1} \
         -padx 9 -pady 3 -text [intlmsg {Add new column}]
    button $page_columns.buttons.rencolbtn \
        -borderwidth 1 \
        -command {
            if {[set PgAcVar(tblinfo,col_id) [$PgAcVar(tblinfo,cols) curselection]]==""} then {
                bell
            } else {
                set PgAcVar(tblinfo,old_cn) [lindex [$PgAcVar(tblinfo,cols) get [$PgAcVar(tblinfo,cols) curselection]] 0]
                set PgAcVar(tblinfo,new_cn) $PgAcVar(tblinfo,old_cn)
                Window show .pgaw:RenameField
            }
        } \
        -padx 9 -pady 3 -text [intlmsg {Rename column}]
    button $page_columns.buttons.addidxbtn \
        -borderwidth 1 -command Tables::addNewIndex \
        -padx 9 -pady 3 -text [intlmsg {Add new index}]

    pack $page_columns.buttons -side bottom -fill x
    grid $page_columns.buttons.addcolbtn \
        -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $page_columns.buttons.rencolbtn \
        -column 1 -row 0 -columnspan 1 -rowspan 1 
    grid $page_columns.buttons.addidxbtn \
        -column 2 -row 0 -columnspan 1 -rowspan 1 
    pack $page_columns.xscroll -side bottom -fill x
    pack $page_columns.yscroll -side right -fill y
    pack $page_columns.cols -side right -anchor nw -expand 1 -fill both

    # indexes page
    scrollbar $page_indexes.xscroll \
        -command [list $page_indexes.indexes xview] \
        -orient horizontal

    scrollbar $page_indexes.yscroll \
        -command [list $page_indexes.indexes yview]

    set PgAcVar(tblinfo,indexes) [tablelist::tablelist $page_indexes.indexes \
        -yscrollcommand [list $page_indexes.yscroll set] \
        -xscrollcommand [list $page_indexes.xscroll set] \
        -labelcommand tablelist::sortByColumn \
        -selectbackground #DDDDDD \
        -selectmode single \
        -font $PgAcVar(pref,font_normal) \
        -labelfont PGfont \
        -stretch all \
        -columns [list \
            0 [intlmsg "indexes"] left \
            0 [intlmsg "index columns"] left \
            0 [intlmsg "is unique?"] left \
            0 [intlmsg "is clustered?"] left \
        ] \
        -selectforeground #708090
    ]
    frame $page_indexes.buttons
    button $page_indexes.buttons.delidxbtn \
        -borderwidth 1 -command Tables::deleteIndex \
        -padx 9 -pady 3 -text [intlmsg {Delete index}]
    button $page_indexes.buttons.clusterbtn \
        -borderwidth 1 -command Tables::clusterIndex \
         -padx 9 -pady 3 -text [intlmsg {Cluster index}]

    pack $page_indexes.buttons -side bottom -fill x
    grid $page_indexes.buttons.delidxbtn \
        -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $page_indexes.buttons.clusterbtn \
        -column 1 -row 0 -columnspan 1 -rowspan 1 
    pack $page_indexes.xscroll -side bottom -fill x
    pack $page_indexes.yscroll -side right -fill y
    pack $page_indexes.indexes -side right -anchor nw -expand 1 -fill both

    # indexes page
    scrollbar $page_permissions.xscroll \
        -command [list $page_permissions.permissions xview] \
        -orient horizontal

    scrollbar $page_permissions.yscroll \
        -command [list $page_permissions.permissions yview]

    set PgAcVar(tblinfo,perms) [tablelist::tablelist $page_permissions.permissions \
        -yscrollcommand [list $page_permissions.yscroll set] \
        -xscrollcommand [list $page_permissions.xscroll set] \
        -labelcommand tablelist::sortByColumn \
        -selectbackground #DDDDDD \
        -selectmode single \
        -font $PgAcVar(pref,font_normal) \
        -labelfont PGfont \
        -stretch all \
        -columns [list \
            0 [intlmsg "user name"] left \
            0 [intlmsg "select"] left \
            0 [intlmsg "update"] left \
            0 [intlmsg "insert"] left \
            0 [intlmsg "rule"] left \
        ] \
        -selectforeground #708090
    ]
    frame $page_permissions.buttons
    button $page_permissions.buttons.adduserbtn \
        -borderwidth 1 -command {
            PgAcVar:clean permission,*
            Window show .pgaw:Permissions
        } \
        -padx 9 -pady 3 -text [intlmsg {Add user}]
    button $page_permissions.buttons.chguserbtn \
        -command Tables::loadPermissions \
        -borderwidth 1 -padx 9 -pady 3 -text [intlmsg {Change permissions}]
    button $page_permissions.buttons.usergrpbtn \
        -command {
            ::Usergroups::designPerms [::Database::quoteObject $PgAcVar(tblinfo,tablename)] table
        } \
        -borderwidth 1 -padx 9 -pady 3 -text [intlmsg {User-Group manager}]

    pack $page_permissions.buttons -side bottom -fill x
    grid $page_permissions.buttons.adduserbtn \
        -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $page_permissions.buttons.chguserbtn \
        -column 1 -row 0 -columnspan 1 -rowspan 1 
    grid $page_permissions.buttons.usergrpbtn \
        -column 2 -row 0 -columnspan 1 -rowspan 1 
    pack $page_permissions.xscroll -side bottom -fill x
    pack $page_permissions.yscroll -side right -fill y
    pack $page_permissions.permissions -side right -anchor nw -expand 1 -fill both

    $base.nb raise [$base.nb pages 0]
    ::tk::PlaceWindow $base widget .

    return
}

proc vTclWindow.pgaw:AddField {base} {
global PgAcVar
    if {$base == ""} {
        set base .pgaw:AddField
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    toplevel $base -class Toplevel
    wm focusmodel $base passive
    wm overrideredirect $base 0
    wm resizable $base 0 0
    wm deiconify $base
    wm title $base [intlmsg "Add new column"]

    # some helpful key bindings
    bind $base <Control-Key-w> [subst {destroy $base}]

    LabelEntry $base.e1 \
        -label [intlmsg {Field name}] \
        -labeljustify right \
        -labelfont $PgAcVar(pref,font_normal) \
        -font $PgAcVar(pref,font_normal) \
        -textvariable PgAcVar(addfield,name)

    LabelEntry $base.e2 \
        -label [intlmsg {Field type}] \
        -labeljustify right \
        -labelfont $PgAcVar(pref,font_normal) \
        -font $PgAcVar(pref,font_normal) \
        -textvariable PgAcVar(addfield,type) \
        -command Tables::addNewColumn

    frame $base.fb
    button $base.fb.b1 \
        -borderwidth 1 -command Tables::addNewColumn -text [intlmsg {Add field}]
    button $base.fb.b2 \
        -borderwidth 1 -command {Window destroy .pgaw:AddField} -text [intlmsg Cancel]

    pack $base.e1 $base.e2 -side top -anchor "e"
    pack $base.fb -side top
    grid $base.fb.b1 -column 0 -row 0
    grid $base.fb.b2 -column 1 -row 0
}

proc vTclWindow.pgaw:AddIndex {base} {
global PgAcVar
    if {$base == ""} {
        set base .pgaw:AddIndex
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    toplevel $base -class Toplevel
    wm focusmodel $base passive
    wm geometry $base 334x203+265+266
    wm maxsize $base 1280 1024
    wm minsize $base 1 1
    wm overrideredirect $base 0
    wm resizable $base 0 0
    wm deiconify $base
    wm title $base [intlmsg "Add new index"]

    # some helpful key bindings
    bind $base <Control-Key-w> [subst {destroy $base}]

    frame $base.f \
        -borderwidth 2 -height 75 -relief groove -width 125 
    frame $base.f.fin \
        -height 75 -relief groove -width 125 
    label $base.f.fin.lin \
        -borderwidth 0 -relief raised -text [intlmsg {Index name}]
    entry $base.f.fin.ein \
        -width 28 -textvariable PgAcVar(addindex,indexname) 
    checkbutton $base.f.cbunique -borderwidth 1 \
        -offvalue { } -onvalue unique -text [intlmsg {Is unique ?}] -variable PgAcVar(addindex,unique)
    label $base.f.ls1 \
        -anchor w -background #dfdbdf -borderwidth 0 -foreground #000086 \
        -justify left -relief raised -textvariable PgAcVar(addindex,indexfields) \
        -wraplength 300 
    label $base.f.lif \
        -borderwidth 0 -relief raised -text "[intlmsg {Index fields}]:"
    label $base.f.ls2 \
        -borderwidth 0 -relief raised -text { } 
    label $base.f.ls3 \
        -borderwidth 0 -relief raised -text { } 
    frame $base.fb \
        -height 75 -relief groove -width 125 
    button $base.fb.btncreate -command Tables::createNewIndex \
        -padx 9 -pady 3 -text [intlmsg Create]
    button $base.fb.btncancel \
        -command {Window destroy .pgaw:AddIndex} -padx 9 -pady 3 -text [intlmsg Cancel]
    pack $base.f \
        -in .pgaw:AddIndex -anchor center -expand 1 -fill both -side top 
    grid $base.f.fin \
        -in .pgaw:AddIndex.f -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $base.f.fin.lin \
        -in .pgaw:AddIndex.f.fin -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $base.f.fin.ein \
        -in .pgaw:AddIndex.f.fin -column 1 -row 0 -columnspan 1 -rowspan 1 
    grid $base.f.cbunique \
        -in .pgaw:AddIndex.f -column 0 -row 5 -columnspan 1 -rowspan 1 
    grid $base.f.ls1 \
        -in .pgaw:AddIndex.f -column 0 -row 3 -columnspan 1 -rowspan 1 
    grid $base.f.lif \
        -in .pgaw:AddIndex.f -column 0 -row 2 -columnspan 1 -rowspan 1 -sticky w 
    grid $base.f.ls2 \
        -in .pgaw:AddIndex.f -column 0 -row 1 -columnspan 1 -rowspan 1 
    grid $base.f.ls3 \
        -in .pgaw:AddIndex.f -column 0 -row 4 -columnspan 1 -rowspan 1 
    pack $base.fb \
        -in .pgaw:AddIndex -anchor center -expand 0 -fill x -side bottom 
    grid $base.fb.btncreate \
        -in .pgaw:AddIndex.fb -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $base.fb.btncancel \
        -in .pgaw:AddIndex.fb -column 1 -row 0 -columnspan 1 -rowspan 1 
}

proc vTclWindow.pgaw:RenameField {base} {
global PgAcVar
    if {$base == ""} {
        set base .pgaw:RenameField
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    toplevel $base -class Toplevel
    wm focusmodel $base passive
    wm geometry $base 215x75+258+213
    wm maxsize $base 1280 1024
    wm minsize $base 1 1
    wm overrideredirect $base 0
    wm resizable $base 0 0
    wm deiconify $base
    wm title $base [intlmsg "Rename column"]

    # some helpful key bindings
    bind $base <Control-Key-w> [subst {destroy $base}]

    label $base.l1 \
        -borderwidth 0 -text [intlmsg {New name}]
    entry $base.e1 \
        -textvariable PgAcVar(tblinfo,new_cn)
    focus $base.e1
    bind $base.e1 <Key-KP_Enter> "Tables::renameColumn"
    bind $base.e1 <Key-Return> "Tables::renameColumn"
    frame $base.f \
        -height 75 -relief groove -width 147 
    button $base.f.b1 \
        -borderwidth 1 -command Tables::renameColumn -text [intlmsg Rename]
    button $base.f.b2 \
        -borderwidth 1 -command {Window destroy .pgaw:RenameField} -text [intlmsg Cancel]
    label $base.l2 -borderwidth 0 
    grid $base.l1 \
        -in .pgaw:RenameField -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $base.e1 \
        -in .pgaw:RenameField -column 1 -row 0 -columnspan 1 -rowspan 1 
    grid $base.f \
        -in .pgaw:RenameField -column 0 -row 4 -columnspan 2 -rowspan 1 
    grid $base.f.b1 \
        -in .pgaw:RenameField.f -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $base.f.b2 \
        -in .pgaw:RenameField.f -column 1 -row 0 -columnspan 1 -rowspan 1 
    grid $base.l2 \
        -in .pgaw:RenameField -column 0 -row 3 -columnspan 1 -rowspan 1 
}

proc vTclWindow.pgaw:Permissions {base} {
    if {$base == ""} {
        set base .pgaw:Permissions
    }
    if {[winfo exists $base]} {
        wm deiconify $base; return
    }
    toplevel $base -class Toplevel
    wm focusmodel $base passive
    wm geometry $base 273x147+256+266
    wm maxsize $base 1280 1024
    wm minsize $base 1 1
    wm overrideredirect $base 0
    wm resizable $base 0 0
    wm deiconify $base
    wm title $base [intlmsg "Permissions"]

    # some helpful key bindings
    bind $base <Control-Key-w> [subst {destroy $base}]

    frame $base.f1 \
        -height 103 -relief groove -width 125 
    label $base.f1.l \
        -borderwidth 0 -relief raised -text [intlmsg {User name}]
    entry $base.f1.ename -textvariable PgAcVar(permission,username)
        
    label $base.f1.l2 \
        -borderwidth 0 -relief raised -text { } 
    label $base.f1.l3 \
        -borderwidth 0 -relief raised -text { } 
    frame $base.f2 \
        -height 75 -relief groove -borderwidth 2 -width 125 
    checkbutton $base.f2.cb1 -borderwidth 1 -padx 4 -pady 4 \
        -text [intlmsg select] -variable PgAcVar(permission,select) 
    checkbutton $base.f2.cb2 -borderwidth 1 -padx 4 -pady 4 \
        -text [intlmsg update] -variable PgAcVar(permission,update)
    checkbutton $base.f2.cb3 -borderwidth 1 -padx 4 -pady 4 \
        -text [intlmsg insert] -variable PgAcVar(permission,insert)
    checkbutton $base.f2.cb4 -borderwidth 1 -padx 4 -pady 4 \
        -text [intlmsg rule] -variable PgAcVar(permission,rule)
    frame $base.fb \
        -height 75 -relief groove -width 125 
    button $base.fb.btnsave -command Tables::savePermissions \
        -padx 9 -pady 3 -text [intlmsg Save]
    button $base.fb.btncancel -command {Window destroy .pgaw:Permissions} \
        -padx 9 -pady 3 -text [intlmsg Close]
    pack $base.f1 \
        -in .pgaw:Permissions -anchor center -expand 0 -fill none -side top 
    grid $base.f1.l \
        -in .pgaw:Permissions.f1 -column 0 -row 1 -columnspan 1 -rowspan 1 
    grid $base.f1.ename \
        -in .pgaw:Permissions.f1 -column 1 -row 1 -columnspan 1 -rowspan 1 -padx 2 
    grid $base.f1.l2 \
        -in .pgaw:Permissions.f1 -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $base.f1.l3 \
        -in .pgaw:Permissions.f1 -column 0 -row 2 -columnspan 1 -rowspan 1 
    pack $base.f2 \
        -in .pgaw:Permissions -anchor center -expand 0 -fill none -side top 
    grid $base.f2.cb1 \
        -in .pgaw:Permissions.f2 -column 0 -row 1 -columnspan 1 -rowspan 1 -sticky w 
    grid $base.f2.cb2 \
        -in .pgaw:Permissions.f2 -column 1 -row 1 -columnspan 1 -rowspan 1 -sticky w 
    grid $base.f2.cb3 \
        -in .pgaw:Permissions.f2 -column 0 -row 2 -columnspan 1 -rowspan 1 -sticky w 
    grid $base.f2.cb4 \
        -in .pgaw:Permissions.f2 -column 1 -row 2 -columnspan 1 -rowspan 1 -sticky w 
    pack $base.fb \
        -in .pgaw:Permissions -anchor center -expand 0 -fill none -pady 3 -side bottom 
    grid $base.fb.btnsave \
        -in .pgaw:Permissions.fb -column 0 -row 0 -columnspan 1 -rowspan 1 
    grid $base.fb.btncancel \
        -in .pgaw:Permissions.fb -column 1 -row 0 -columnspan 1 -rowspan 1 

    focus $base.f1.ename
}

