#==========================================================
# Triggers --
#
#   handling of PostgreSQL triggers
#   triggers will be referenced like so in PGA:
#       schema.table.trigger_name
#==========================================================
#
namespace eval Triggers {
    variable rowstatement
    variable beforeafter
    variable eventslist
}


#----------------------------------------------------------
# ::Triggers::new --
#
#   sets up to create a new Trigger
#
# Arguments:
#   none
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Triggers::new {} {

    showError [intlmsg "Not yet implemented."]

}; # end proc ::Triggers::new


#----------------------------------------------------------
# ::Triggers::open --
#
#   passes work to design proc for opening a Trigger
#
# Arguments:
#   trigger_    name of Trigger to open in design mode
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Triggers::open {trigger_} {

    ::Triggers::design $trigger_

}; # end proc ::Triggers::open


#----------------------------------------------------------
# ::Triggers::design --
#
#   opens window in design mode for selected Trigger
#
# Arguments:
#   trigger_    name of Trigger to open in design mode
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Triggers::design {trigger_} {

    showError [intlmsg "Not yet implemented."]

}; # end proc ::Triggers::design


#----------------------------------------------------------
# ::Triggers::introspect --
#
#   Given a triggername, returns the SQL needed to recreate it
#
# Arguments:
#   triggername_    name of a trigger to introspect
#   dbh_            an optional database handle
#
# Returns:
#   insql      the CREATE statement to make this trigger
#----------------------------------------------------------
#
proc ::Triggers::introspect {triggername_ {dbh_ ""}} {

    set insql [::Triggers::clone $triggername_ $triggername_ $dbh_]

    return $insql

}; # end proc ::Triggers::introspect


#----------------------------------------------------------
# ::Triggers::clone --
#
#   Like introspect, only changes the triggername
#
# Arguments:
#   srctrigger_     the original trigger
#   desttrigger_    the clone trigger
#   dbh_            an optional database handle
#
# Returns:
#   insql       the CREATE statement to clone this trigger
#----------------------------------------------------------
#
proc ::Triggers::clone {srctrigger_ desttrigger_ {dbh_ ""}} {

    global CurrentDB

    variable rowstatement
    variable beforeafter
    variable eventslist

    if {[string match "" $dbh_]} {
        set dbh_ $CurrentDB
    }

    set V [::Database::getPgVersion $dbh_]

    set insql ""

    set srctrigger_ [string map {\" {}} $srctrigger_]
    set desttrigger_ [string map {\" {}} $desttrigger_]

    set srctbl [join [lrange [split $srctrigger_ .] 0 end-1] .]
    set srctrig [lindex [split $srctrigger_ .] end]
    set dsttbl [join [lrange [split $desttrigger_ .] 0 end-1] .]
    set dsttrig [lindex [split $desttrigger_ .] end]

    # check in case we accidentally deleted the dst tbl name
    if {[string length $dsttbl]==0} {
        set dsttbl $srctbl
    }

    set sql ""

    if {$V<7.3} {
        set sql "SELECT t.*, p.proname
                   FROM pg_trigger t,
                        pg_proc p,
                        pg_class c
                  WHERE t.tgname='$srctrig'
                    AND t.tgrelid=c.oid
                    AND t.tgfoid=p.oid
                    AND c.relname='$srctbl'"
    } else {
        set nsp [lindex [split $srctbl .] 0]
        set srctbl [lindex [split $srctbl .] 1]
        set sql "SELECT t.*, p.proname
                   FROM pg_catalog.pg_trigger t,
                        pg_catalog.pg_proc p,
                        pg_catalog.pg_class c,
                        pg_catalog.pg_namespace n
                  WHERE t.tgname='$srctrig'
                    AND t.tgrelid=c.oid
                    AND t.tgfoid=p.oid
                    AND c.relname='$srctbl'
                    AND c.relnamespace=n.oid
                    AND n.nspname='$nsp'"
    }

    wpg_select $dbh_ $sql rec {
        breakDownBitField $rec(tgtype)
        set targs [string trim [string map {\\000 ,} $rec(tgargs)] ,]
        set insql "CREATE TRIGGER [::Database::quoteObject $dsttrig]
\t$beforeafter
\t[join $eventslist " OR "]
\tON [::Database::quoteObject $dsttbl]
\tFOR EACH $rowstatement
\tEXECUTE PROCEDURE $rec(proname) ($targs)
;"
    }

    return $insql

}; # end proc ::Triggers::clone



#----------------------------------------------------------
# ::Triggers::breakDownBitField --
#
#   figures out some nitty-gritty details about Triggers
#   by setting several internal namespace variables
#
# Arguments:
#   bits_   a bit field to decipher
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Triggers::breakDownBitField {bits_} {

    variable rowstatement
    variable beforeafter
    variable eventslist

    set trow [expr {1 << 0}]
    set tbefore [expr {1 << 1}]
    set tinsert [expr {1 << 2}]
    set tdelete [expr {1 << 3}]
    set tupdate [expr {1 << 4}]

    if {[expr {$bits_ & $trow}]} {
        set rowstatement "ROW"
    } else {
        set rowstatement "STATEMENT"
    }

    if {[expr {$bits_ & $tbefore}]} {
        set beforeafter "BEFORE"
    } else {
        set beforeafter "AFTER"
    }

    set eventslist [list]
    if {[expr {$bits_ & $tinsert}]} {
        lappend eventslist "INSERT"
    }
    if {[expr {$bits_ & $tdelete}]} {
        lappend eventslist "DELETE"
    }
    if {[expr {$bits_ & $tupdate}]} {
        lappend eventslist "UPDATE"
    }

}; # end proc ::Triggers::breakDownBitField



# from PG source, include/catalog/pg_trigger.h

#define TRIGGER_TYPE_ROW>--->--->--->---(1 << 0)
#define TRIGGER_TYPE_BEFORE>>--->--->---(1 << 1)
#define TRIGGER_TYPE_INSERT>>--->--->---(1 << 2)
#define TRIGGER_TYPE_DELETE>>--->--->---(1 << 3)
#define TRIGGER_TYPE_UPDATE>>--->--->---(1 << 4)

#define TRIGGER_CLEAR_TYPE(type)>--->---(type = 0)

#define TRIGGER_SETT_ROW(type)>->--->---(type |= TRIGGER_TYPE_ROW)
#define TRIGGER_SETT_BEFORE(type)>-->---(type |= TRIGGER_TYPE_BEFORE)
#define TRIGGER_SETT_INSERT(type)>-->---(type |= TRIGGER_TYPE_INSERT)
#define TRIGGER_SETT_DELETE(type)>-->---(type |= TRIGGER_TYPE_DELETE)
#define TRIGGER_SETT_UPDATE(type)>-->---(type |= TRIGGER_TYPE_UPDATE)

#define TRIGGER_FOR_ROW(type)>-->--->---(type & TRIGGER_TYPE_ROW)
#define TRIGGER_FOR_BEFORE(type)>--->---(type & TRIGGER_TYPE_BEFORE)
#define TRIGGER_FOR_INSERT(type)>--->---(type & TRIGGER_TYPE_INSERT)
#define TRIGGER_FOR_DELETE(type)>--->---(type & TRIGGER_TYPE_DELETE)
#define TRIGGER_FOR_UPDATE(type)>--->---(type & TRIGGER_TYPE_UPDATE)

