#==========================================================
# Types --
#
#   handling of PostgreSQL types
#==========================================================
#
namespace eval Types {}


#----------------------------------------------------------
# ::Types::init --
#
#   clears out the namespace variables and opens the window
#   for Types
#
# Arguments:
#   none
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::init {} {

    variable Win
    variable mode
    variable typename ""
    variable typetype "composite"
    variable inputfcn ""
    variable outputfcn ""
    variable recvfcn ""
    variable sendfcn ""
    variable internallength ""
    variable passedby ""
    variable alignment ""
    variable storage ""
    variable defaultval ""
    variable elementtypes ""
    variable delimiter ""
    variable comment

}; # end proc ::Types init


#----------------------------------------------------------
# ::Types::new --
#
#   sets up to create a new type
#
# Arguments:
#   none
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::new {} {

    variable Win
    variable mode

    ::Types::init

    set mode "new"

    Window show .pgaw:Type

    $Win(page,composite-comment) delete 1.0 end
    $Win(page,base-comment) delete 1.0 end

}; # end proc ::Types::new


#----------------------------------------------------------
# ::Types::open --
#
#   passes work to design proc for opening a Type
#
# Arguments:
#   typename_   name of type to open in design mode
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::open {typename_} {

    ::Types::design $typename_

}; # end proc ::Types::open


#----------------------------------------------------------
# ::Types::design --
#
#   opens a Type in design mode
#
# Arguments:
#   typename_   name of type to design
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::design {typename_} {

    variable Win
    variable mode
    variable typename
    variable typetype
    variable comment

    ::Types::init

    set mode "design"
    set typename $typename_
    # this will set typetype and comment
    getTypeOfType $typename_

    Window show .pgaw:Type

    switch -- $typetype {
        composite {
            $Win(notebook) delete base
            $Win(notebook) delete pseudo
            foreach pair [getCompositePairsList $typename_] {
                $Win(page,composite).table insert end $pair
            }
            $Win(page,composite-comment) delete 1.0 end
            $Win(page,composite-comment) insert end $comment
        }
        base {
            $Win(notebook) delete composite
            $Win(notebook) delete pseudo
            loadBaseType $typename_
            $Win(page,base-comment) delete 1.0 end
            $Win(page,base-comment) insert end $comment
        }
        default {}
    }

}; # end proc ::Types::design


#----------------------------------------------------------
# ::Types::introspect --
#
#   Given a typename, returns the SQL needed to recreate it
#
# Arguments:
#   typename_   name of a type to introspect
#   dbh_        an optional database handle
#
# Returns:
#   insql       the CREATE statement to make this type
#----------------------------------------------------------
#
proc ::Types::introspect {typename_ {dbh_ ""}} {

    set insql [::Types::clone $typename_ $typename_ $dbh_]

    return $insql

}; # end proc ::Types::introspect


#----------------------------------------------------------
# ::Types::clone --
#
#   Like introspect, only changes the typename.  Relies on
#   code from PG source in pg_dump.c to figure out the
#   CREATE statement.  Much easier than re-inventing the
#   wheel.
#
# Arguments:
#   srctype_    the original type
#   desttype_   the clone type
#   dbh_        an optional database handle
#
# Returns:
#   insql       the CREATE statement to clone this type
#----------------------------------------------------------
#
proc ::Types::clone {srctype_ desttype_ {dbh_ ""}} {

    global CurrentDB
    variable typetype

    if {[string match "" $dbh_]} {
        set dbh_ $CurrentDB
    }

    set V [::Database::getPgVersion $dbh_]

    set insql ""

    set srctype_ [string map {\" {}} $srctype_]
    set srcnsp [lindex [split $srctype_ .] 0]
    set srctyp [lindex [split $srctype_ .] 1]
    set desttype_ [string map {\" {}} $desttype_]

    # note the sql produced below borrows heavily from pg_dump.c
    set sql ""

    # this is redundant, since getTypeOfType already sets
    # typetype, but just in case it changes sometime
    set typetype [::Types::getTypeOfType $srctype_]

    # this is a little different from other clone procs,
    # which normally just work at version numbers, but there
    # are 3 different types, 2 of which we care about here
    if {$typetype=="composite"} {
        if {$V>=7.3} {
            set sql "SELECT a.attname,
                            pg_catalog.format_type(a.atttypid, a.atttypmod)
                         AS atttypdefn
                       FROM pg_catalog.pg_type t,
                            pg_catalog.pg_attribute a,
                            pg_catalog.pg_namespace n
                      WHERE a.attrelid=t.typrelid
                        AND NOT a.attisdropped
                        AND t.typname='$srctyp'
                        AND t.typnamespace=n.oid
                        AND n.nspname='$srcnsp'"
            set insql "CREATE TYPE [::Database::quoteObject $desttype_] ("
            wpg_select $dbh_ $sql rec {
                append insql "\n\t$rec(attname) $rec(atttypdefn)"
            }
            append insql "\n);"
        } else {
            # we eventually need to support types better
            # pre-7.3
        }

    } elseif {$typetype=="base"} {
        # base types will take much, much more work
        if {$V>=7.4} {
            set sql "SELECT t.typlen,
                            t.typinput, t.typoutput, t.typreceive, t.typsend,
                            t.typinput::pg_catalog.oid AS typinputoid,
                            t.typoutput::pg_catalog.oid AS typoutputoid,
                            t.typreceive::pg_catalog.oid AS typreceiveoid,
                            t.typsend::pg_catalog.oid AS typsendoid,
                            t.typdelim, t.typdefault, t.typbyval, t.typalign,
                            CASE WHEN t.typalign='c' THEN 'char'
                                 WHEN t.typalign='s' THEN 'int2'
                                 WHEN t.typalign='i' THEN 'int4'
                                 WHEN t.typalign='d' THEN 'double'
                                 ELSE ''
                            END AS alignment,
                            CASE WHEN t.typstorage='p' THEN 'plain'
                                 WHEN t.typstorage='e' THEN 'external'
                                 WHEN t.typstorage='x' THEN 'extended'
                                 WHEN t.typstorage='m' THEN 'main'
                                 ELSE ''
                            END AS storage,
                            t.typstorage
                       FROM pg_catalog.pg_type t,
                            pg_catalog.pg_namespace n
                      WHERE t.typname='$srctyp'
                        AND t.typnamespace=n.oid
                        AND n.nspname='$srcnsp'"
        } elseif {$V>=7.3} {
            set sql "SELECT t.typlen,
                            t.typinput, t.typoutput,
                            '-' AS typreceive,
                            '-' AS typsend,
                            t.typinput::pg_catalog.oid AS typinputoid,
                            t.typoutput::pg_catalog.oid AS typoutputoid,
                            0 AS typreceiveoid,
                            0 AS typsendoid,
                            t.typdelim, t.typdefault, t.typbyval, t.typalign,
                            CASE WHEN t.typalign='c' THEN 'char'
                                 WHEN t.typalign='s' THEN 'int2'
                                 WHEN t.typalign='i' THEN 'int4'
                                 WHEN t.typalign='d' THEN 'double'
                                 ELSE ''
                            END AS alignment,
                            CASE WHEN t.typstorage='p' THEN 'plain'
                                 WHEN t.typstorage='e' THEN 'external'
                                 WHEN t.typstorage='x' THEN 'extended'
                                 WHEN t.typstorage='m' THEN 'main'
                                 ELSE ''
                            END AS storage,
                            t.typstorage
                       FROM pg_catalog.pg_type t,
                            pg_catalog.pg_namespace n
                      WHERE t.typname='$srctyp'
                        AND t.typnamespace=n.oid
                        AND n.nspname='$srcnsp'"
        # we are going way back, how nice of pg_dump to show
        # us the way :)
        } elseif {$V>=7.1} {
            set sql "SELECT t.typlen,
                            t.typinput, t.typoutput,
                            '-' AS typreceive,
                            '-' AS typsend,
                            t.typinput::oid AS typinputoid,
                            t.typoutput::oid AS typoutputoid,
                            0 AS typreceiveoid,
                            0 AS typsendoid,
                            t.typdelim, t.typdefault, t.typbyval, t.typalign,
                            CASE WHEN t.typalign='c' THEN 'char'
                                 WHEN t.typalign='s' THEN 'int2'
                                 WHEN t.typalign='i' THEN 'int4'
                                 WHEN t.typalign='d' THEN 'double'
                                 ELSE ''
                            END AS alignment,
                            CASE WHEN t.typstorage='p' THEN 'plain'
                                 WHEN t.typstorage='e' THEN 'external'
                                 WHEN t.typstorage='x' THEN 'extended'
                                 WHEN t.typstorage='m' THEN 'main'
                                 ELSE ''
                            END AS storage,
                            t.typstorage
                       FROM pg_type t
                      WHERE t.typname='$srctype_'"
        } else {
            set sql "SELECT t.typlen,
                            t.typinput, t.typoutput,
                            '-' AS typreceive,
                            '-' AS typsend,
                            t.typinput::oid AS typinputoid,
                            t.typoutput::oid AS typoutputoid,
                            0 AS typreceiveoid,
                            0 AS typsendoid,
                            t.typdelim, t.typdefault, t.typbyval, t.typalign,
                            CASE WHEN t.typalign='c' THEN 'char'
                                 WHEN t.typalign='s' THEN 'int2'
                                 WHEN t.typalign='i' THEN 'int4'
                                 WHEN t.typalign='d' THEN 'double'
                                 ELSE ''
                            END AS alignment,
                            'plain' AS storage
                       FROM pg_type t
                      WHERE t.typname='$srctype_'"
        }
        wpg_select $dbh_ $sql rec {
            set insql "CREATE TYPE [::Database::quoteObject $desttype_] ("
            append insql "\n\tINTERNALLENGTH="
            if {$rec(typlen)==-1} {
                append insql "variable,"
            } else {
                append insql $rec(typlen) ","
            }
            append insql "\n\tINPUT=[::Database::quoteObject $rec(typinput)],"
            append insql "\n\tOUTPUT=[::Database::quoteObject $rec(typoutput)],"
            if {$rec(typreceiveoid)!=0} {
                append insql "\n\tRECEIVE=[::Database::quoteObject $rec(typreceive)],"
            }
            if {$rec(typsendoid)!=0} {
                append insql "\n\tSEND=[::Database::quoteObject $rec(typsend)],"
            }
            if {[string length $rec(typdefault)]>0} {
                append insql "\n\tDEFAULT=\"$rec(typdefault)\","
            }
            if {$rec(typdelim)!=","} {
                append insql "\n\tDELIMITER=\"$rec(delim)\","
            }
            append insql "\n\tALIGNMENT=$rec(alignment),"
            append insql "\n\tSTORAGE=$rec(storage),"
            if {$rec(typbyval)=="t"} {
                append insql "\n\tPASSEDBYVALUE"
            }
            set insql [string trimright $insql ","]
            append insql "\n);"
        }

    }

    return $insql

}; # end proc ::Types::clone


#----------------------------------------------------------
# ::Types::loadBaseType --
#
#   loads a base type for design mode
#
# Arguments:
#   typename_   name of type to load
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::loadBaseType {typename_} {

    global CurrentDB
    variable inputfcn
    variable outputfcn
    variable recvfcn
    variable sendfcn
    variable internallength
    variable passedby
    variable alignment
    variable storage
    variable defaultval
    variable elementtypes
    variable delimiter

    set sql ""

    # if we are schema qualified, strip out the schema name
    set schema [string range $typename_ 0 [expr {[string first . $typename_]-1}]]
    if { [string length $schema] > 0 } {
        set typename_ [string range $typename_ [expr {[string length $schema]+1}] end]
    }

    set V [::Database::getPgVersion $CurrentDB]

    if {$V < 7.3} {
        # dont really care about 7.2 anymore
        # upgrade your databases, people
        showError [intlmsg "Operation not supported on PostgreSQL versions before 7.3, sorry"]
        return
    } else {
        # mmm, big tasty sql...
        set sql "
            SELECT t.typlen,
                   CASE WHEN t.typlen=-1 THEN 'VARIABLE'
                        WHEN t.typlen=-2 THEN 'cstring'
                        ELSE t.typlen::text
                   END AS internallength,
                   t.typbyval,
                   CASE WHEN t.typbyval='t' THEN 'VALUE'
                        ELSE 'REFERENCE'
                   END AS passedby,
                   t.typalign,
                   CASE WHEN t.typalign='c' THEN 'char'
                        WHEN t.typalign='s' THEN 'int2'
                        WHEN t.typalign='i' THEN 'int4'
                        WHEN t.typalign='d' THEN 'double'
                        ELSE ''
                   END AS alignment,
                   t.typstorage,
                   CASE WHEN t.typstorage='p' THEN 'plain'
                        WHEN t.typstorage='e' THEN 'external'
                        WHEN t.typstorage='x' THEN 'extended'
                        WHEN t.typstorage='m' THEN 'main'
                        ELSE ''
                   END AS storage,
                   t.typdefault,
                   COALESCE(t.typdefault,'NULL') AS defaultval,
                   t.typdelim,
                   t.typdelim AS delimiter,
                   t.typelem,
                   (SELECT e.typname
                      FROM pg_catalog.pg_type e
                     WHERE e.oid=t.typelem) AS elementtypes,
                   t.typinput,
                   (SELECT s.nspname || '.' || p.proname
                      FROM pg_catalog.pg_proc p,
                           pg_catalog.pg_namespace s
                     WHERE p.oid=t.typinput
                       AND p.pronamespace=s.oid) AS inputfcn,
                   t.typoutput,
                   (SELECT s.nspname || '.' || p.proname
                      FROM pg_catalog.pg_proc p,
                           pg_catalog.pg_namespace s
                     WHERE p.oid=t.typoutput
                       AND p.pronamespace=s.oid) AS outputfcn"

            if {$V >= 7.4} {
                append sql ",
                   t.typreceive,
                   (SELECT s.nspname || '.' || p.proname
                      FROM pg_catalog.pg_proc p,
                           pg_catalog.pg_namespace s
                     WHERE p.oid=t.typreceive
                       AND p.pronamespace=s.oid) AS recvfcn,
                   t.typsend,
                   (SELECT s.nspname || '.' || p.proname
                      FROM pg_catalog.pg_proc p,
                           pg_catalog.pg_namespace s
                     WHERE p.oid=t.typsend
                       AND p.pronamespace=s.oid) AS sendfcn"
            }; # end 7.4 version check

            append sql "
              FROM pg_catalog.pg_type t,
                   pg_catalog.pg_namespace n
             WHERE t.typname='$typename_'
               AND t.typnamespace=n.oid
               AND n.nspname='$schema'"
    }

    if {[catch {wpg_select $CurrentDB "$sql" rec {
            set internallength $rec(internallength)
            set passedby $rec(passedby)
            set storage $rec(storage)
            set alignment $rec(alignment)
            set defaultval $rec(defaultval)
            set delimiter $rec(delimiter)
            set elementtypes $rec(elementtypes)
            set inputfcn $rec(inputfcn)
            set outputfcn $rec(outputfcn)
            if {$V >= 7.4} {
                set recvfcn $rec(recvfcn)
                set sendfcn $rec(sendfcn)
            }; # end 7.4 version check
        }
    } err]} {
        showError $err
    }

}; # end proc ::Types::loadBaseType


#----------------------------------------------------------
# ::Types::addCompositePair --
#
#   inserts an item into the table for an
#   attribute_name/type pair
#
# Arguments:
#   none
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::addCompositePair {} {

    variable Win
    variable mode

    if {$mode == "design"} {
        set size [expr {[$Win(page,composite).table size]+1}]
        $Win(page,composite).table insert end [list $size "DEFAULT" "char"]
    } else {
        $Win(page,composite).table insert end [list "DEFAULT" "char"]
    }

}; # end proc ::Types::addCompositePair


#----------------------------------------------------------
# ::Types::removeCompositePair --
#
#   deletes an attribute_name/type pair from the list
#
# Arguments:
#   none
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::removeCompositePair {} {

    variable Win

    set cursel [$Win(page,composite).table curselection]

    if {$cursel == ""} {return ""}

    $Win(page,composite).table delete $cursel

}; # end proc ::Types::removeCompositePair


#----------------------------------------------------------
# ::Types::changeCompositeType --
#
#   alters the type of an attribute_name/type pair
#
# Arguments:
#   newtype_    the new type to assign to the pair
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::changeCompositeType {newtype_} {

    variable Win
    variable mode

    set cursel [$Win(page,composite).table curselection]

    if {$cursel == ""} {return ""}

    if {$mode == "design"} {
        append cursel ",2"
    } else {
        append cursel ",1"
    }

    $Win(page,composite).table cellconfigure $cursel \
        -text $newtype_

}; # end proc ::Types::changeCompositeType


#----------------------------------------------------------
# ::Types::saveCompositeType --
#
#   create the type in the database
#
# Arguments:
#   saveas_     boolean indicating whether or not to save
#               this type separately or overwrite it
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::saveCompositeType {{saveas_ 0}} {

    global CurrentDB

    variable Win
    variable mode
    variable typename

    if { [string length $typename] == 0 } {
        showError [intlmsg "You must supply a name for the Type!"]
        return
    }

    if {$mode == "design"} {
        $Win(page,composite).table sortbycolumn 0
    }

    set sql "BEGIN TRANSACTION"
    sql_exec noquiet $sql

    if {!$saveas_ && $mode=="design"} {
        set sql "DROP TYPE $typename"
        sql_exec noquiet $sql
    }

    set sql "CREATE TYPE $typename AS "

    set pairlist [list]
    foreach pair [$Win(page,composite).table get 0 end] {
        if {$mode == "design"} {
            lappend pairlist [lrange $pair 1 end]
        } else {
            lappend pairlist $pair
        }
    }

    append sql "(" [join $pairlist ,] ")"

    if {[sql_exec noquiet $sql]} {
        set sql "COMMIT TRANSACTION"
        sql_exec noquiet $sql
    } else {
        set sql "ROLLBACK TRANSACTION"
        sql_exec noquiet $sql
    }

    if {[string length [string trim [join $pairlist]]]==0} {
        showError [intlmsg "You must supply more information!"]
        return
    }

    saveComment

    ::Mainlib::tab_click Types

}; # end proc ::Types::saveCompositeType


#----------------------------------------------------------
# ::Types::saveComment --
#
#   saves the comment on a type
#
# Arguments:
#   delete_     if 1, sets comment string to NULL,
#               thus deleting the comment. default 0
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::saveComment {{delete_ 0}} {

    variable Win
    variable typename
    variable typetype

    set sql "COMMENT ON TYPE [::Database::quoteObject $typename]"

    if {$delete_} {
        append sql " IS NULL"
        $Win(page,$typetype-comment) delete 1.0 end
    } else {
        append sql " IS '[string trim [$Win(page,$typetype-comment) get 1.0 end]]'"
    }

    sql_exec noquiet $sql

}; # end proc ::Types::saveComment


#----------------------------------------------------------
# ::Types::saveBaseType --
#
#   create the type in the database
#
# Arguments:
#   saveas_     boolean indicating whether or not to save
#               this type separately or overwrite it
#
# Returns:
#   none
#----------------------------------------------------------
#
proc ::Types::saveBaseType {{saveas_ 0}} {

    global CurrentDB

    variable Win
    variable mode
    variable typename
    variable inputfcn
    variable outputfcn
    variable recvfcn
    variable sendfcn
    variable internallength
    variable passedby
    variable alignment
    variable storage
    variable defaultval
    variable elementtypes
    variable delimiter

    if { [string length $typename] == 0 } {
        showError [intlmsg "You must supply a name for the Type!"]
        return
    }

    if { [string length $inputfcn] == 0 } {
        showError [intlmsg "You must supply an input function for the Type!"]
        return
    }

    if { [string length $outputfcn] == 0 } {
        showError [intlmsg "You must supply an output function for the Type!"]
        return
    }

    set sql "BEGIN TRANSACTION"
    sql_exec noquiet $sql

    if {!$saveas_ && $mode=="design"} {
        set sql "DROP TYPE $typename"
        sql_exec noquiet $sql
    }

    set sql "CREATE TYPE $typename ("

    set p1 [string last ( $inputfcn]
    set p2 [string last ) $inputfcn]
    set ifcn [string replace $inputfcn $p1 $p2]
    append sql "INPUT=$ifcn,"

    set p1 [string last ( $outputfcn]
    set p2 [string last ) $outputfcn]
    set ofcn [string replace $outputfcn $p1 $p2]
    append sql "OUTPUT=$ofcn"

    if {[string length $recvfcn] > 0} {
        set p1 [string last ( $recvfcn]
        set p2 [string last ) $recvfcn]
        set rfcn [string replace $recvfcn $p1 $p2]
        append sql ", RECEIVE=$rfcn"
    }

    if {[string length $sendfcn] > 0} {
        set p1 [string last ( $sendfcn]
        set p2 [string last ) $sendfcn]
        set sfcn [string replace $sendfcn $p1 $p2]
        append sql ", SEND=$sfcn"
    }

    if {[string length $internallength] > 0} {
        append sql ", INTERNALLENGTH=$internallength"
    }

    if {$passedby==[intlmsg "VALUE"]} {
        append sql ", PASSEDBYVALUE"
    }

    if {[string length $alignment] > 0} {
        append sql ", ALIGNMENT=$alignment"
    }

    if {[string length $storage] > 0} {
        append sql ", STORAGE=$storage"
    }

    if {$defaultval!="NULL"} {
        append sql ", DEFAULT='$defaultval'"
    }

    if {[string length $elementtypes] > 0} {
        append sql ", ELEMENT=$elementtypes"
    }

    if {[string length $delimiter] > 0} {
        append sql ", DELIMITER='$delimiter'"
    }

    append sql ")"

    if {[sql_exec noquiet $sql]} {
        set sql "COMMIT TRANSACTION"
        sql_exec noquiet $sql
        saveComment
        ::Mainlib::tab_click Types
    } else {
        set sql "ROLLBACK TRANSACTION"
        sql_exec noquiet $sql
    }

}; # end proc ::Types::saveBaseType


#----------------------------------------------------------
# ::Types::getTypeOfType --
#
#   Not only will this return the name of the type's type
#   that we supply as a parameter, it will set the namespace
#   typetype variable to this value.
#
# Arguments:
#   typename_   name of the type we are taking a look at.
#
# Results:
#   typ         the type of the type
#
# Modifies:
#   typetype    the namespace var
#----------------------------------------------------------
#
proc ::Types::getTypeOfType {typename_} {

    global CurrentDB

    variable typetype
    variable comment

    set sql ""
    set typ ""
    set cmt ""

    # if we are schema qualified, strip out the schema name
    set schema [string range $typename_ 0 [expr {[string first . $typename_]-1}]]
    if { [string length $schema] > 0 } {
        set typename_ [string range $typename_ [expr {[string length $schema]+1}] end]
    }

    set V [::Database::getPgVersion $CurrentDB]

    if {$V < 7.3} {
        set sql "
            SELECT CASE WHEN t.typtype='b' THEN 'base'
                        WHEN t.typtype='c' THEN 'composite'
                        WHEN t.typtype='d' THEN 'domain'
                        WHEN t.typtype='p' THEN 'pseudo'
                        ELSE 'other'
                   END AS typtype,
                   obj_description(t.oid) AS tdesc
              FROM pg_type t
             WHERE t.typname='$typename_'"
    } else {
        set sql "
            SELECT CASE WHEN t.typtype='b' THEN 'base'
                        WHEN t.typtype='c' THEN 'composite'
                        WHEN t.typtype='d' THEN 'domain'
                        WHEN t.typtype='p' THEN 'pseudo'
                        ELSE 'other'
                   END AS typtype,
                   obj_description(t.oid) AS tdesc
              FROM pg_catalog.pg_type t,
                   pg_catalog.pg_namespace n
             WHERE t.typname='$typename_'
               AND t.typnamespace=n.oid
               AND n.nspname='$schema'"
    }

    if {[catch {wpg_select $CurrentDB "$sql" rec {
            set typ $rec(typtype)
            set cmt $rec(tdesc)
        }
    } err]} {
        showError $err
    }

    set typetype $typ
    set comment $cmt

    return $typ

}; # end proc ::Types::getTypeOfType


#----------------------------------------------------------
#----------------------------------------------------------
#
proc ::Types::getCompositePairsList {typename_} {

    global CurrentDB

    set sql ""

    # if we are schema qualified, strip out the schema name
    set schema [string range $typename_ 0 [expr {[string first . $typename_]-1}]]
    if { [string length $schema] > 0 } {
        set typename_ [string range $typename_ [expr {[string length $schema]+1}] end]
    }

    set V [::Database::getPgVersion $CurrentDB]

    if {$V < 7.3} {
        showError [intlmsg "Operation not supported on PostgreSQL versions before 7.3, sorry"]
        return
    } else {
        set sql "
            SELECT DISTINCT
                   a.attnum, a.attname,
                   (SELECT tt.typname
                      FROM pg_type tt
                     WHERE tt.oid=a.atttypid)
                   AS typtype
              FROM pg_attribute a,
                   pg_catalog.pg_type t,
                   pg_catalog.pg_namespace n
             WHERE t.typname='$typename_'
               AND a.attrelid=t.typrelid
               AND t.typnamespace=n.oid
               AND n.nspname='$schema'
          ORDER BY a.attnum"
    }

    set pairs [list]
    if {[catch {wpg_select $CurrentDB "$sql" rec {
            lappend pairs [list $rec(attnum) $rec(attname) $rec(typtype)]
        }
    } err]} {
        showError $err
    }

    return $pairs

}; # end proc ::Types::getCompositePairsList



### END TYPES NAMESPACE ###
### BEGIN VISUAL TCL CODE ###



proc vTclWindow.pgaw:Type {base} {

    global PgAcVar
    global CurrentDB

    if { [string length $base] == 0 } {
        set base .pgaw:Type
    }

    if {[winfo exists $base]} {
        wm deiconify $base
        catch {$::Types::Win(notebook) raise $::Types::typetype}
        return
    }

    toplevel $base -class Toplevel
    wm focusmodel $base passive
    wm geometry $base 500x400+100+40
    wm overrideredirect $base 0
    wm resizable $base 1 1
    wm deiconify $base
    wm title $base [intlmsg "Type"]

    # some helpful key bindings
    bind $base <Control-Key-w> [subst {destroy $base}]
    bind $base <Key-F1> "Help::load types"

    set V [::Database::getPgVersion $CurrentDB]

    set ::Types::Win(notebook) [NoteBook $base.nb]
    pack $::Types::Win(notebook) \
        -expand 1 \
        -fill both

    set ::Types::Win(page,composite) \
        [$::Types::Win(notebook) insert end composite \
            -text [intlmsg "Composite"]]

    set ::Types::Win(page,base) \
        [$::Types::Win(notebook) insert end base \
            -text [intlmsg "Base"]]

    set ::Types::Win(page,pseudo) \
        [$::Types::Win(notebook) insert end pseudo \
            -text [intlmsg "Pseudo"]]

    #
    # composite page
    #

    LabelEntry $::Types::Win(page,composite).lename \
        -label [intlmsg "Name"] \
        -textvariable ::Types::typename

    scrollbar $::Types::Win(page,composite).xscroll \
        -width 12 \
        -command [list $::Types::Win(page,composite).table xview] \
        -highlightthickness 0 \
        -orient horizontal \
        -background #DDDDDD \
        -takefocus 0

    scrollbar $::Types::Win(page,composite).yscroll \
        -width 12 \
        -command [list $::Types::Win(page,composite).table yview] \
        -highlightthickness 0 \
        -background #DDDDDD \
        -takefocus 0

    tablelist::tablelist $::Types::Win(page,composite).table \
        -yscrollcommand \
            [list $::Types::Win(page,composite).yscroll set] \
        -xscrollcommand \
            [list $::Types::Win(page,composite).xscroll set] \
        -labelcommand tablelist::sortByColumn \
        -stripebg #e0e8f0 \
        -selectbackground #DDDDDD \
        -selectmode extended \
        -font $PgAcVar(pref,font_normal) \
        -labelfont $PgAcVar(pref,font_bold) \
        -stretch all \
        -selectforeground #708090 \
        -labelbackground #DDDDDD \
        -labelforeground navy

    set compcols [list]
    if {$::Types::mode == "design"} {
        lappend compcols 0 [intlmsg "order"] left
    }
    lappend compcols \
        0 [intlmsg "attribute name"] left \
        0 [intlmsg "data type"] left

    $::Types::Win(page,composite).table configure \
        -columns $compcols

    # dictionary sort is the best of ascii and integer

    for {set i 0} {$i < [expr {[llength $compcols]/3}]} {incr i} {
        $::Types::Win(page,composite).table columnconfigure $i \
            -sortmode dictionary
    }
    for {set i 0} {$i < [expr {[llength $compcols]/3-1}]} {incr i} {
        $::Types::Win(page,composite).table columnconfigure $i \
            -editable 1
    }

    # buttons to add/remove a composite attribute/type pair

    ButtonBox $::Types::Win(page,composite).bboxtable \
        -orient vertical \
        -homogeneous 1 \
        -spacing 2
    $::Types::Win(page,composite).bboxtable add \
        -image ::icon::hotlistadd-16 \
        -helptext [intlmsg "Add"] \
        -borderwidth 1 \
        -command {::Types::addCompositePair}
    $::Types::Win(page,composite).bboxtable add \
        -image ::icon::hotlistdel-16 \
        -helptext [intlmsg "Remove"] \
        -borderwidth 1 \
        -command {::Types::removeCompositePair}

    # a popup for the different base types

    menu $::Types::Win(page,composite).table.typepopup
    set newcol 0
    foreach basetype [join [::Database::getTypesList {} 1 0 0 "b" 0]] {
        $::Types::Win(page,composite).table.typepopup add command \
            -label $basetype \
            -command "::Types::changeCompositeType $basetype" \
            -columnbreak [expr {floor((($newcol%20)+1)/20)}]
        incr newcol
    }
    set body [$::Types::Win(page,composite).table bodypath]
    bind $body <ButtonRelease-3> {
        tk_popup $::Types::Win(page,composite).table.typepopup %X %Y 0
    }

    # some buttons to save and whatnot

    ButtonBox $::Types::Win(page,composite).bboxmenu \
        -orient horizontal \
        -homogeneous 1 \
        -spacing 2
    $::Types::Win(page,composite).bboxmenu add \
        -relief link \
        -image ::icon::filesave-22 \
        -helptext [intlmsg "Save"] \
        -borderwidth 1 \
        -command {::Types::saveCompositeType 0}
    $::Types::Win(page,composite).bboxmenu add \
        -relief link \
        -image ::icon::filesaveas-22 \
        -helptext [intlmsg "Save As"] \
        -borderwidth 1 \
        -command {::Types::saveCompositeType 1}
    $::Types::Win(page,composite).bboxmenu add \
        -relief link \
        -image ::icon::help-22 \
        -helptext [intlmsg "Help"] \
        -borderwidth 1 \
        -command {::Help::load types}
    $::Types::Win(page,composite).bboxmenu add \
        -relief link \
        -image ::icon::exit-22 \
        -helptext [intlmsg "Close"] \
        -borderwidth 1 \
        -command {
            catch {Window destroy .pgaw:Type}
        }

    # comment
    TitleFrame $::Types::Win(page,composite).comment \
        -text [intlmsg "Comments"] \
        -font $PgAcVar(pref,font_bold)
    set titleframe [$::Types::Win(page,composite).comment getframe]
    set ::Types::Win(page,composite-comment) $titleframe.text
    text $titleframe.text \
        -height 3 \
        -width 3
    ButtonBox $titleframe.btnbox \
        -orient vertical \
        -homogeneous 1
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::filesave-22 \
        -helptext [intlmsg "Save"] \
        -borderwidth 1 \
        -command {::Types::saveComment 0}
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::edittrash-22 \
        -helptext [intlmsg "Delete"] \
        -borderwidth 1 \
        -command {::Types::saveComment 1}
    grid $titleframe.text \
        -row 0 \
        -column 0 \
        -sticky news
    grid $titleframe.btnbox \
        -row 0 \
        -column 1 \
        -sticky e
    grid columnconfigure $titleframe 0 \
        -weight 10

    grid $::Types::Win(page,composite).lename \
        -row 0 \
        -column 0 \
        -columnspan 2 \
        -sticky we
    grid $::Types::Win(page,composite).bboxmenu \
        -row 0 \
        -column 2 \
        -columnspan 3 \
        -sticky e
    grid $::Types::Win(page,composite).table \
        -row 1 \
        -column 0 \
        -rowspan 2 \
        -columnspan 3 \
        -sticky news
    grid $::Types::Win(page,composite).yscroll \
        -row 1 \
        -column 3 \
        -rowspan 2 \
        -sticky swn
    grid $::Types::Win(page,composite).xscroll \
        -row 3 \
        -column 0 \
        -columnspan 3 \
        -sticky wen
    grid $::Types::Win(page,composite).bboxtable \
        -padx 5 \
        -ipadx 5 \
        -pady 20 \
        -ipady 20 \
        -row 2 \
        -column 4 \
        -sticky n
    grid $::Types::Win(page,composite).comment \
        -row 4 \
        -column 0 \
        -columnspan 3 \
        -sticky sew

    grid columnconfigure $::Types::Win(page,composite) 1 \
        -weight 10
    grid rowconfigure $::Types::Win(page,composite) 2 \
        -weight 10

    #
    # basic page
    #

    Label $::Types::Win(page,base).lname \
        -text [intlmsg "Name (*)"]
    Entry $::Types::Win(page,base).ename \
        -textvariable ::Types::typename

    set bpitems [list]

    lappend bpitems [list linputfcn cbinputfcn]
    Label $::Types::Win(page,base).linputfcn \
        -text [intlmsg "Input Function (*)"]
    ComboBox $::Types::Win(page,base).cbinputfcn \
        -textvariable ::Types::inputfcn \
        -values [::Database::getFunctionsList] \
        -editable false

    lappend bpitems [list loutputfcn cboutputfcn]
    Label $::Types::Win(page,base).loutputfcn \
        -text [intlmsg "Output Function (*)"]
    ComboBox $::Types::Win(page,base).cboutputfcn \
        -textvariable ::Types::outputfcn \
        -values [::Database::getFunctionsList] \
        -editable false

    if {$V >= 7.4} {

        lappend bpitems [list lrecvfcn cbrecvfcn]
        Label $::Types::Win(page,base).lrecvfcn \
            -text [intlmsg "Receive Function"]
        ComboBox $::Types::Win(page,base).cbrecvfcn \
            -textvariable ::Types::recvfcn \
            -values [::Database::getFunctionsList] \
            -editable true

        lappend bpitems [list lsendfcn cbsendfcn]
        Label $::Types::Win(page,base).lsendfcn \
            -text [intlmsg "Send Function"]
        ComboBox $::Types::Win(page,base).cbsendfcn \
            -textvariable ::Types::sendfcn \
            -values [::Database::getFunctionsList] \
            -editable true

    }; # end 7.4 version check

    lappend bpitems [list linternallength cbinternallength]
    Label $::Types::Win(page,base).linternallength \
        -text [intlmsg "Internal Length"]
    ComboBox $::Types::Win(page,base).cbinternallength \
        -textvariable ::Types::internallength \
        -text [intlmsg "VARIABLE"] \
        -values [list [intlmsg "VARIABLE"] 1 2 3 4 5 6 7 8 9 ...] \
        -editable true

    lappend bpitems [list lpassedby cbpassedby]
    Label $::Types::Win(page,base).lpassedby \
        -text [intlmsg "Passed By"]
    ComboBox $::Types::Win(page,base).cbpassedby \
        -textvariable ::Types::passedby \
        -text [intlmsg "REFERENCE"] \
        -values [list [intlmsg "VALUE"] [intlmsg "REFERENCE"]] \
        -editable false

    lappend bpitems [list lalignment cbalignment]
    Label $::Types::Win(page,base).lalignment \
        -text [intlmsg "Alignment"]
    ComboBox $::Types::Win(page,base).cbalignment \
        -textvariable ::Types::alignment \
        -text int4 \
        -values [list char int2 int4 double] \
        -editable false

    lappend bpitems [list lstorage cbstorage]
    Label $::Types::Win(page,base).lstorage \
        -text [intlmsg "Storage"]
    ComboBox $::Types::Win(page,base).cbstorage \
        -textvariable ::Types::storage \
        -text plain \
        -values [list plain external extended main] \
        -editable false

    lappend bpitems [list ldefaultval edefaultval]
    Label $::Types::Win(page,base).ldefaultval \
        -text [intlmsg "Default"]
    Entry $::Types::Win(page,base).edefaultval \
        -text NULL \
        -textvariable ::Types::defaultval

    lappend bpitems [list lelement eelement]
    Label $::Types::Win(page,base).lelement \
        -text [intlmsg "Element Types (if an array)"]
    Entry $::Types::Win(page,base).eelement \
        -textvariable ::Types::elementtypes

    lappend bpitems [list ldelim edelim]
    Label $::Types::Win(page,base).ldelim \
        -text [intlmsg "Delimiter"]
    Entry $::Types::Win(page,base).edelim \
        -text , \
        -textvariable ::Types::delimiter

    # some buttons to save and whatnot

    ButtonBox $::Types::Win(page,base).bboxmenu \
        -orient horizontal \
        -homogeneous 1 \
        -spacing 2
    $::Types::Win(page,base).bboxmenu add \
        -relief link \
        -image ::icon::filesave-22 \
        -helptext [intlmsg "Save"] \
        -borderwidth 1 \
        -command {::Types::saveBaseType 0}
    $::Types::Win(page,base).bboxmenu add \
        -relief link \
        -image ::icon::filesaveas-22 \
        -helptext [intlmsg "Save As"] \
        -borderwidth 1 \
        -command {::Types::saveBaseType 1}
    $::Types::Win(page,base).bboxmenu add \
        -relief link \
        -image ::icon::help-22 \
        -helptext [intlmsg "Help"] \
        -borderwidth 1 \
        -command {::Help::load types}
    $::Types::Win(page,base).bboxmenu add \
        -relief link \
        -image ::icon::exit-22 \
        -helptext [intlmsg "Close"] \
        -borderwidth 1 \
        -command {
            catch {Window destroy .pgaw:Type}
        }

    # comment
    TitleFrame $::Types::Win(page,base).comment \
        -text [intlmsg "Comments"] \
        -font $PgAcVar(pref,font_bold)
    set titleframe [$::Types::Win(page,base).comment getframe]
    set ::Types::Win(page,base-comment) $titleframe.text
    text $titleframe.text \
        -height 3 \
        -width 3
    ButtonBox $titleframe.btnbox \
        -orient vertical \
        -homogeneous 1 \
        -spacing 0
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::filesave-22 \
        -helptext [intlmsg "Save"] \
        -borderwidth 1 \
        -command {::Types::saveComment 0}
    $titleframe.btnbox add \
        -relief link \
        -image ::icon::edittrash-22 \
        -helptext [intlmsg "Delete"] \
        -borderwidth 1 \
        -command {::Types::saveComment 1}
    grid $titleframe.text \
        -row 0 \
        -column 0 \
        -sticky news
    grid $titleframe.btnbox \
        -row 0 \
        -column 1 \
        -sticky e
    grid columnconfigure $titleframe 0 \
        -weight 10

    grid $::Types::Win(page,base).lname \
        -row 0 \
        -column 0 \
        -columnspan 1 \
        -sticky e
    grid $::Types::Win(page,base).ename \
        -row 0 \
        -column 1 \
        -columnspan 1 \
        -sticky we
    grid $::Types::Win(page,base).bboxmenu \
        -row 0 \
        -column 2 \
        -columnspan 3 \
        -sticky e

    # i am so lazy
    set row 1
    foreach item $bpitems {
        set litem [lindex $item 0]
        set ritem [lindex $item 1]
        grid $::Types::Win(page,base).$litem \
            -row $row \
            -column 0 \
            -columnspan 1 \
            -sticky e \
            -pady 2
        grid $::Types::Win(page,base).$ritem \
            -row $row \
            -column 1 \
            -columnspan 1 \
            -sticky ew
        incr row
    }

    incr row
    grid $::Types::Win(page,base).comment \
        -row $row \
        -column 0 \
        -columnspan 3 \
        -sticky sew

    grid columnconfigure $::Types::Win(page,base) 1 \
        -weight 10
    grid columnconfigure $::Types::Win(page,base) 2 \
        -weight 10
    grid rowconfigure $::Types::Win(page,base) $row \
        -weight 10

    #
    # pseudo page
    #
    label $::Types::Win(page,pseudo).lbl \
        -text [intlmsg "Not yet implemented!  Should it be?  Hmmm."]
    pack $::Types::Win(page,pseudo).lbl

    # dont just show a blank screen
    $::Types::Win(notebook) raise $::Types::typetype

}; # end proc vTclWindow.pgaw:Type


