#----------------------------------------------------------
# cmd_line.tcl
#
#    Parsing the command line
#
#----------------------------------------------------------

debug "\nParsing the command line parameters:"

# loading the cmdline tcl package (part of the tcllib package)
if {[catch {package require cmdline} msg]} {
	debug "\tCouldn't find cmdline package elsewhere on your machine."
	debug "\tNow trying to load the cmdline package from PgAccess."
	lappend ::auto_path [file join $PgAcVar(PGACCESS_HOME) lib] [file join $PgAcVar(PGACCESS_HOME) lib widgets]
	if {[catch {
		debug "\t"
		debug "\tcmdline: [package require cmdline]"
		debug "\tGot cmdline package supplied by PgAccess !"
		debug "\t"
	} msg]} {
		# error message and exit if package loading failed
		if {[lsearch argv "-debug"] != -1} {
			set PgAcVar(debug) true
		}
		debug "\tFAILED"
		puts stderr "Error: $msg"
		puts stderr "Please install the tcllib package (http:://www.sourceforge.net/projects/tcllib)"
		exit
	}
}

# the command line parameters defined as an
#
# array: {name [list]} where
#
#   name: the parameter
#   list: [type description action] where
#
#     type: with (.arg) or without ("") argument
#     description: string to display if help needed
#     action: the action to take if this parameter is set.
#             The $val variable gives the value specified for that parameter.

array set cmdline_args {
	dir {
		.arg "Path to configuration file directory" {
			set PgAcVar(PGACCESS_DIR) $val
			debug "\tPGACCESS_DIR: $PgAcVar(PGACCESS_DIR)"
		}
	}
	pglib {
		.arg "Path to libpgtcl shared object (dll/so) file" {
			set PgAcVar(PGLIB) $val
			debug "\tPGLIB: $PgAcVar(PGLIB)"
		}
	}
	home {
		.arg "PGACCESS_HOME" {
			set PgAcVar(PGACCESS_HOME) $val
			debug "\tPGACCESS_HOME: $PgAcVar(PGACCESS_HOME)"
		}
	}
	version {
		"" "Show version information" {
			puts "\tPGACCESS_VERSION: $PgAcVar(VERSION)"
			exit 0
		}
	}
	debug {
		"" "Show debug information" {
			set PgAcVar(DEBUG) true
			debug "\tDEBUG: $PgAcVar(DEBUG)"
		}
	}
	hide {
		"" "Hide main window (will also disable the splash screen)" {
			set PgAcVar(PGACCESS_HIDDEN) 1
			debug "\tPGACCESS_HIDDEN: $PgAcVar(PGACCESS_HIDDEN)"
			# disable the splash screen too
			set PgAcVar(PGACCESS_NOSPLASH) 1
			debug "\tPGACCESS_NOSPLASH: $PgAcVar(PGACCESS_NOSPLASH)"
		}
	}
	nosplash {
		"" "Disable splash screen" {
			set PgAcVar(PGACCESS_NOSPLASH) 1
			debug "\tPGACCESS_NOSPLASH: $PgAcVar(PGACCESS_NOSPLASH)"
		}
	}
	noauto {
		"" "Disables automatic load of connections otherwise set to auto-load at startup" {
			set PgAcVar(PGACCESS_NOAUTO) 1
			debug "\tPGACCESS_NOAUTO: $PgAcVar(PGACCESS_NOAUTO)"
		}
	}
	noconn {
		"" "Disables loading and saving of connections file" {
			set PgAcVar(PGACCESS_NOCONN) 1
			debug "\tPGACCESS_NOCONN: $PgAcVar(PGACCESS_NOCONN)"
		}
	}
	noscript {
		"" "Disables execution of the 'autoexec' script" {
			set PgAcVar(PGACCESS_NOSCRIPT) 1
			debug "\tPGACCESS_NOSCRIPT: $PgAcVar(PGACCESS_NOSCRIPT)"
		}
	}
	fast {
		"" "Disables pga tables check, splash screen, auto db connect, and auto exec scripts." {
			set PgAcVar(PGACCESS_NOPGATABLESINIT) 1
			debug "\tPGACCESS_NOPGATABLESINIT: $PgAcVar(PGACCESS_NOPGATABLESINIT)"
			set PgAcVar(PGACCESS_NOSCRIPT) 1
			debug "\tPGACCESS_NOSCRIPT: $PgAcVar(PGACCESS_NOSCRIPT)"
			set PgAcVar(PGACCESS_NOAUTO) 1
			debug "\tPGACCESS_NOAUTO: $PgAcVar(PGACCESS_NOAUTO)"
			set PgAcVar(PGACCESS_NOSPLASH) 1
			debug "\tPGACCESS_NOSPLASH: $PgAcVar(PGACCESS_NOSPLASH)"
		}
	}
	
	conn {
		"" "Automatically loads the connection specified on the command line" {
			set PgAcVar(PGACCESS_CONNLOAD) 1
			debug "\tPGACCESS_CONNLOAD: $PgAcVar(PGACCESS_CONNLOAD)"
		}
	}
	login {
		"" "Displays login dialog at startup" {
			set PgAcVar(PGACCESS_LOGIN) 1
			debug "\tPGACCESS_LOGIN: $PgAcVar(PGACCESS_LOGIN)"
		}
	}
	host {
		.arg "Database host" {
			set PgAcVar(opendb,host) $val
			debug "\topendb,host: $PgAcVar(opendb,host)"
		}
	}
	pgport {
		.arg "Database port" {
			set PgAcVar(opendb,pgport) $val
			debug "\topendb,pgport: $PgAcVar(opendb,pgport)"
		}
	}
	dbname {
		.arg "Database name" {
			set PgAcVar(opendb,dbname) $val
			debug "\topendb,dbname: $PgAcVar(opendb,dbname)"
		}
	}
	username {
		.arg "Database user" {
			set PgAcVar(opendb,username) $val
			debug "\topendb,username: $PgAcVar(opendb,username)"
		}
	}
	password {
		.arg "Database password" {
			set PgAcVar(opendb,password) $val
			debug "\topendb,password: $PgAcVar(opendb,password)"
		}
	}
	pgintcl {
		"" "Uses pgin.tcl frontend/backend protocol v3 (for PG>=7.4), a pure Tcl-PG interface, instead of a dll/so (overrides PGLIB)" {
			set PgAcVar(PGINTCL) 1
			debug "\tPGINTCL: $PgAcVar(PGINTCL)"
		}
	}
	temp {
		"" "Creates only TEMPorary PgAccess tables so your database doesn't get cluttered by them." {
			set PgAcVar(PGACCESS_TEMP) 1
			debug "\tPGACCESS_TEMP: $PgAcVar(PGACCESS_TEMP)"
		}
	}
	smlogin {
		"" "Shrinks the login window to ask for only items not specified on the command line" {
			set PgAcVar(PGACCESS_SMLOGIN) 1
			set PgAcVar(PGACCESS_LOGIN) 1
			debug "\tPGACCESS_SMLOGIN: $PgAcVar(PGACCESS_SMLOGIN)"
			debug "\tPGACCESS_LOGIN: $PgAcVar(PGACCESS_LOGIN)"
		}
	}
	printcmd {
		.arg "Print command to set as default." {
			set PgAcVar(PGACCESS_PRINTCMD) $val
			set PgAcVar(pref,print_cmd) $val
			debug "\tPGACCESS_PRINTCMD: $PgAcVar(PGACCESS_PRINTCMD)"
		}
	}
	nicepreview {
		"" "Makes the print preview window for reports nicer." {
			set PgAcVar(PGACCESS_NICEREPORTPREVIEW) 1
			debug "\tPGACCESS_NICEREPORTPREVIEW: $PgAcVar(PGACCESS_NICEREPORTPREVIEW)"
		}
	}
	
}

# creating the getopt list for the parsing script
foreach i [lsort [array names cmdline_args]] {
	lappend getopt_list "$i[lindex $cmdline_args($i) 0]"
}

# parsing the command line variables and executing (evaluating) the specified script
while {[set ok [cmdline::getopt argv $getopt_list opt val]] > 0} {
	if {[catch {eval [lindex $cmdline_args($opt) 2]} msg]} {
		puts "Error: $msg ..."
	}
}

# in case of error
if {$ok < 0} {
	puts "Error: wrong option"
	foreach i [lsort [array names cmdline_args]] {
		lappend usage_list "$i \{[lindex $cmdline_args($i) 1]\}"
	}
	puts [cmdline::usage $usage_list]
	exit -1
}
