#----------------------------------------------------------
# load.tcl
#
#    Loading plugins and namespaces
#
#----------------------------------------------------------

#----------------------------------------------------------
# init_load_plugins --
#
#    Loads the plugins on startup
#
# Arguments:
#    none
#
# Results:
#    none returned.
#----------------------------------------------------------
proc init_load_plugins {} {
	global PgAcVar
	debug "\nLoading plugins:"
	# loading all the plugins from the plugins directory
	foreach plugin [glob -nocomplain [file join $PgAcVar(PGACCESS_HOME) lib plugins *.tcl]] {
		set plug [file rootname [file tail $plugin]]
		# special check for pgin.tcl
		if { [string match "pgin*" $plug] } {
			if { [info exists PgAcVar(PGINTCL)] } {
				if { [string match "pgin" $plug] } {
					debug "\t$plug"
					source $plugin
				}
			}
		} else {
			debug "\t$plug"
			source $plugin
		}
		
	}
	
}; # end proc init_load_plugins

#----------------------------------------------------------
# registerPlugin --
#
#    Takes care of registering the plugin
#
# Arguments:
#    name_    the name to appear in the plugins menu
#    cmd_     the command to invoke the plugin
#
# Results:
#    none returned.
#----------------------------------------------------------
#
proc registerPlugin {name_ cmd_} {
	if {![info exists ::PgAcVar(plugin,list)]} {
		set ::PgAcVar(plugin,list) [list]
	}
	
	lappend ::PgAcVar(plugin,list) $name_
	set ::PgAcVar(plugin,$name_) $cmd_
}; # end proc registerPlugin

#----------------------------------------------------------
# init_load_namespaces --
#
#    Loads all defined namespaces
#
# Arguments:
#    none
#
# Results:
#    none returned.
#----------------------------------------------------------
#
proc init_load_namespaces {} {
	global PgAcVar
	# Loading all defined namespaces
	debug "\nLoading namespaces:"
	foreach module $PgAcVar(namespaces) {
		debug "\t$module"
		if {[catch {source [file join $PgAcVar(PGACCESS_HOME) lib $module.tcl]} msg]} {
			puts "\nERROR: $msg"
			puts "Please check your installation or set the PGACCESS_HOME ($PgAcVar(PGACCESS_HOME)) variable properly!"
			exit -1
		}
	}
}; # end proc init_load_namespaces

#----------------------------------------------------------
# re_load_namespaces --
#
#    Reloads all defined namespaces
#
# Arguments:
#    none
#
# Results:
#    none returned.
#----------------------------------------------------------
#
proc re_load_namespaces {} {
	global PgAcVar
	# Reloading all defined namespaces
	debug "\nReloading namespaces:"
	set namespaces [namespace children ::]
	foreach module $PgAcVar(namespaces) {
		if {$module!="mainlib"} {
			foreach namespace $namespaces {
				if {[string match -nocase "??$module" $namespace]} {
					namespace delete $namespace
					debug "\t$module"
					if {[catch {source [file join $PgAcVar(PGACCESS_HOME) lib $module.tcl]} msg]} {
						puts "\nERROR: $msg"
						puts "Please check your installation or set the PGACCESS_HOME ($PgAcVar(PGACCESS_HOME)) variable properly!"
						exit -1
					}
				}
			}
		}
	}
}; # end proc re_load_namespaces



#----------------------------------------------------------
# init_load_required_packages --
#
#    Loads the required packages
#
# Arguments:
#    none
#
# Results:
#    none returned.
#----------------------------------------------------------
#
proc init_load_required_packages {} {
	global PgAcVar
	#  Loading the required packages
	debug "\nLoading the required packages:"
	lappend ::auto_path [file join $PgAcVar(PGACCESS_HOME) lib] [file join $PgAcVar(PGACCESS_HOME) lib widgets]
	if {[catch {
		debug "\ttablelist: [package require tablelist 3.8]"
		debug "\tBWidget: [package require BWidget 1.7.0]"
		debug "\ticons: [package require icons]"
		debug "\tbase64: [package require base64]"
		debug "\thtmlparse: [package require htmlparse]"
		debug "\tcsv: [package require csv]"
		debug "\tfileutil: [package require fileutil]"
		debug "\ttkwizard: [package require tkwizard]"
		debug "\tBarcode: [package require Barcode]"
		debug "\tmsgcat: [package require msgcat]"
	} msg]} {
		puts stderr "\nERROR: $msg"
		puts stderr "\n[intlmsg {Please install the required packages:}]\n"
		puts stderr "\ttablelist: http:://www.nemethi.de"
		puts stderr "\ttcllib: http:://www.sourceforge.net/projects/tcllib"
		puts stderr "\tBWidget: http:://www.sourceforge.net/projects/tcllib"
		puts stderr "\ttkwizard: http://www2.clearlight.com/~oakley/tcl/tkwizard/"
		exit -1
	}
	
}; # end proc init_load_required_packages
